"""Compatibility wrapper when major libs changes API."""

from __future__ import annotations

import matplotlib
import scipy

# pylint: disable=no-member


def _version_ge_than(proposed, comparewith):
    """Simple version checker, assuming version on form x.y.z

    Returns True of proposed is >= comparewith
    """

    xvx = proposed.split(".")
    yvx = comparewith.split(".")

    xversion = float(f"{xvx[0]}.{int(xvx[1]):03d}{int(xvx[2]):03d}")
    yversion = float(f"{yvx[0]}.{int(yvx[1]):03d}{int(yvx[2]):03d}")

    return xversion >= yversion


def matplotlib_colormap(name):
    """Compatibility wrapper for matplotlib colormaps syntax that change across 3.6."""
    return (
        matplotlib.colormaps[name]
        if _version_ge_than(matplotlib.__version__, "3.6.0")
        else matplotlib.pyplot.cm.get_cmap(name)
    )


def scipy_gaussianfilter(inputv, sigma):
    """Compatibility wrapper for scipy gaussian_filter syntax that change across 1.8."""
    return (
        scipy.ndimage.gaussian_filter(inputv, sigma)
        if _version_ge_than(scipy.__version__, "1.8.0")
        else scipy.ndimage.filters.gaussian_filter(inputv, sigma)
    )



import os
import pathlib

import pytest
import xtgeo

from xtgeoviz.plot import Grid3DSlice

TPATH = pathlib.Path("../xtgeo-testdata")

USEFILE1 = TPATH / "3dgrids/reek/reek_sim_grid.roff"
USEFILE2 = TPATH / "3dgrids/reek/reek_sim_poro.roff"
USEFILE3 = TPATH / "etc/colortables/rainbow_reverse.rmscolor"


def test_slice_simple_layer(tmpdir, show_plot, generate_plot):
    """Trigger XSection class, and do some simple things basically."""
    layslice = Grid3DSlice()

    mygrid = xtgeo.grid_from_file(USEFILE1)
    myprop = xtgeo.gridproperty_from_file(USEFILE2, grid=mygrid, name="PORO")

    assert myprop.values.mean() == pytest.approx(0.1677, abs=0.001)

    wd = None  # [457000, 464000, 1650, 1800]
    for lay in range(1, mygrid.nlay + 1):
        layslice.canvas(title=f"My Grid Layer plot for layer {lay}")
        layslice.plot_gridslice(
            mygrid,
            prop=myprop,
            mode="layer",
            index=lay,
            window=wd,
            linecolor="black",
        )

        if show_plot:
            layslice.show()
        if generate_plot:
            layslice.savefig(os.path.join(tmpdir, "layerslice_" + str(lay) + ".png"))
        layslice.close()



#!/usr/bin/env python3

import os
import sys
from datetime import date

import xtgeoviz

cwd = os.getcwd()
project_root = os.path.dirname(cwd) + "/src"
sys.path.insert(0, project_root)


# -- General configuration ---------------------------------------------

# The full version, including alpha/beta/rc tags.
release = xtgeoviz.__version__

extensions = [
    "myst_parser",
    "sphinxcontrib.apidoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
]

apidoc_module_dir = "../src/xtgeoviz"
apidoc_output_dir = "apiref"
apidoc_excluded_paths = ["tests"]
apidoc_separate_modules = True
apidoc_module_first = True
apidoc_extra_args = ["-H", "API reference for xtgeoviz"]

autoclass_content = "both"

napoleon_include_special_with_doc = False

# The suffix of source filenames.
source_suffix = {".rst": "restructuredtext", ".md": "markdown"}

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "xtgeoviz"
current_year = date.today().year
copyright = "Equinor 2019 - " + str(current_year) + f" (XTGeo release {release})"


# Sort members by input order in classes
autodoc_member_order = "bysource"
autodoc_default_flags = ["members", "show_inheritance"]

exclude_patterns = ["_build"]

pygments_style = "sphinx"

html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "style_nav_header_background": "#C0C0C0",
}


# html_logo = "images/xtgeo-logo.png"

# The name of an image file (within the static path) to use as favicon
# of the docs.  This file should be a Windows icon file (.ico) being
# 16x16 or 32x32 pixels large.
# html_favicon = None


# If not '', a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names
# to template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer.
# Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer.
# Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages
# will contain a <link> tag referring to it.  The value of this option
# must be the base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "xtgeoviz"



import os
import pathlib

import pytest


def pytest_configure(config):
    import matplotlib as mpl

    mpl.use("Agg")


def pytest_addoption(parser):
    parser.addoption(
        "--testdatapath",
        help="path to xtgeo-testdata, defaults to ../xtgeo-testdata"
        "and is overriden by the XTG_TESTPATH environment variable."
        "Experimental feature, not all tests obey this option.",
        action="store",
        default="../xtgeo-testdata",
    )
    parser.addoption(
        "--generate-plots",
        help="whether to generate plot files. The files are written to the"
        "pytest tmpfolder. In order to inspect whether plots are correctly"
        "generated, the files must be manually inspected.",
        action="store_true",
        default=False,
    )


@pytest.fixture(name="generate_plot")
def fixture_generate_plot(request):
    if "ROXENV" in os.environ:
        pytest.skip("Skip plotting tests in roxar environment")
    return request.config.getoption("--generate-plots")


@pytest.fixture(name="show_plot")
def fixture_xtgshow():
    """For eventual plotting, to be uses in an if sence inside a test."""
    if "ROXENV" in os.environ:
        pytest.skip("Skip plotting tests in roxar environment")
    return any(word in os.environ for word in ["XTGSHOW", "XTG_SHOW"])


@pytest.fixture(name="testdir")
def fixure_testdir():
    """Relative path to xtgeo-testdata, defaulted to ../xtgeo-testdata."""
    relative_path_default = "../xtgeo-testdata"

    return pathlib.Path(os.environ.get("XTGEO_TESTDIR", relative_path_default))



"""Config option for xsections"""

from __future__ import annotations

import logging
import re
from copy import deepcopy
from typing import Optional, Union

import six
import yaml

logger = logging.getLogger(__name__)

DEFAULT_CONFIG = """
title: UnknownField

verbosity: normal  # alt: silent, info, debug

input:

    # source of various xsections
    wells:
        objects: No  # xtgeo Wells() object, alternative to folder/wildcard
        folder: No
        wildcard: "*"
        zonelog: ZONELOG
        perflog: No
        facieslog: No

    # data to show along xsections
    surfaces:
        primary:
            objects: No
            folder: No
            names: []
            wildcard: "*"

        secondary:
            objects: No
            folder: No
            names: []
            wildcard: "*"

        contacts:
            objects: No
            folder: No
            names: []
            wildcard: "*"

    outline: No

    cube: No

    grid:
        surfaces: No
        geometry: No
        property:
            file: No
            name: No
            date: No


plotsettings:

    general:
        engine: matplotlib  # Alternatives not implemented yet

    design:
        style: 1
        zrange: smart1
        gridlines: both
        subtitle: ""
        legendsize: 6
        legends: True
        dpi: 100

    wellmap:
        otherwells: No
        expand: 2

    surfaces:
        primary:
            colors: random40
            fill: Yes

        secondary:
            colors: xtgeo
            legend: Secondary

        contacts:
            colors: rainbow

    cube:
        colors: seismic
        range: []
        interpolation: gaussian
        alpha: 0.8
        sampling: nearest

    grid:
        colors: rainbow
        range: []
        alpha: 0.8
        zinc: 0.5

    wells:
        zonelog:
            colors: random40
            zoneshift: 0
            colordict: {}

        facieslog:
            colors: xtgeo
            colordict: {}

        perflog:
            colors: xtgeo
            colordict: {}

        wellcrossings:
            show: No
            sampling: 20
            wfilter: 5

output:
    plotfolder: /tmp
    format: svg
    prefix: No
    pdfjoin: No
    cleanup: No
"""


class ConfigError(Exception):
    """Exception for config reading error."""


def config_defaults(as_yaml: Optional[bool] = True) -> Union[str, dict]:
    """The YAML file full spec with defaults!

    Args:
        as_yaml: If True, then return as string on YAML format, else, return a
            python dict.
    """
    logger.info("Configure default settings with as_yaml set to: %s", as_yaml)
    if as_yaml:
        return DEFAULT_CONFIG

    return yaml.safe_load(DEFAULT_CONFIG)


def _invalid_key(superset, subset):
    """Simple test if keyword in subset is present in superset."""
    if not isinstance(superset, dict) or not isinstance(subset, dict):
        return ""

    for key, values in subset.items():
        if key not in superset:
            return str(key)

        if isinstance(values, dict) and isinstance(superset[key], dict):
            key = _invalid_key(superset[key], subset[key])

            if key != "":
                return key

    return ""


# Borrowed from https://stackoverflow.com/a/15836901
def data_merge(superset, subset, recursive=False):
    """Merges subset into superset and return merged result

    The `recursive` flag is just to be able to detect the state for logging msg.

    NOTE: tuples and arbitrary objects are not handled as it is totally
    ambiguous what should happen
    """
    if not recursive:
        logger.info("Merging user settings with default settings.")

    checkkey = _invalid_key(superset, subset)
    if checkkey != "":
        raise ValueError(f"Input contains invalid fields: {checkkey}")

    key = None
    try:
        if superset is None or isinstance(
            superset, (six.string_types, six.integer_types, float)
        ):
            # border case for first run or if a is a primitive
            superset = subset
        elif isinstance(superset, list):
            # lists can be only appended
            if isinstance(subset, list):
                # merge lists
                superset.extend(subset)
            else:
                # append to list
                superset.append(subset)
        elif isinstance(superset, dict):
            # dicts must be merged
            if isinstance(subset, dict):
                for key in subset:
                    if key in superset:
                        superset[key] = data_merge(
                            superset[key], subset[key], recursive=True
                        )
                    else:
                        superset[key] = subset[key]
            else:
                raise ConfigError(
                    f'Cannot merge non-dict "{subset}" into dict "{superset}"'
                )
        else:
            raise ConfigError(f'NOT IMPLEMENTED "{subset}" into "{superset}"')
    except TypeError as err:
        raise ConfigError(
            f"Error '{err}' in key '{key}' when merging '{subset}' into '{superset}'"
        )

    return superset


def config_complete(self):
    """Improving the config to make all sorts of defaults and smart stuff."""

    newcfg = deepcopy(self.config)
    oldcfg = self.config

    prm = self.surfaces["primary"]

    # PLOTSETTINGS
    # ----------------------------------------------------------------------------------
    newcfg_plt = newcfg["plotsettings"]
    oldcfg_plt = oldcfg["plotsettings"]

    ddict = {i: i for i in range(100)}  # just a default dict

    if not oldcfg_plt["wells"]["facieslog"]["colordict"]:
        newcfg_plt["wells"]["facieslog"]["colordict"] = ddict

    if not oldcfg_plt["wells"]["perflog"]["colordict"]:
        newcfg_plt["wells"]["perflog"]["colordict"] = ddict

    if not oldcfg_plt["wells"]["zonelog"]["colordict"]:
        newcfg_plt["wells"]["zonelog"]["colordict"] = ddict

    # PROCESSING SMARTIES
    # ----------------------------------------------------------------------------------
    _zrange = deepcopy(newcfg_plt["design"]["zrange"])
    if not _zrange:
        _zrange = "smart0"

    if isinstance(_zrange, str):
        if _zrange == "smart0":
            newcfg_plt["design"]["zrange"] = _config_smart0_zrange(prm)
        elif _zrange == "smart1":
            newcfg_plt["design"]["zrange"] = _config_smart1_zrange(
                self.wells["wlist"], prm
            )
    elif isinstance(_zrange, dict):
        # process zrange per well, with a global default
        newcfg_plt["design"]["zrange"] = _config_userdefined_zrange(
            self.wells["wlist"], _zrange
        )

    elif isinstance(_zrange, list) and len(_zrange) == 2:
        logger.info("Keep zrange as is")

    else:
        raise RuntimeError("Cannot process/understand the zrange for unknown reasons")

    # Z RANGES
    # ----------------------------------------------------------------------------------
    if self.cube and not newcfg_plt["cube"]["range"]:
        newcfg_plt["cube"]["range"] = [self.cube.values.min(), self.cube.values.max()]

    if self.grid and not newcfg_plt["grid"]["range"]:
        newcfg_plt["grid"]["range"] = [
            self.gridproperty.values.min(),
            self.gridproperty.values.max(),
        ]

    self.config = newcfg


def _config_smart0_zrange(primary):
    """If no zrange is given, try to guess from surfaces"""
    minv = primary[0].values.min()
    maxv = primary[-1].values.max()

    minv = round(minv - 10, -1)
    maxv = round(maxv + 10, -1)

    return [minv, maxv]


def _config_smart1_zrange(wlist, primary):
    minv = primary[0].values.min()
    maxv = primary[-1].values.max()

    minv = round(minv - 10, -1)
    maxv = round(maxv + 10, -1)

    maxwell = minv
    for well in wlist:
        zval = well.dataframe["Z_TVDSS"].values
        maxw = zval.max()
        maxw = round(maxw + 10, -1)
        if maxw > maxwell:
            maxwell = maxw

    if maxwell < maxv:
        maxv = maxwell

    return [minv, maxv]


def _config_userdefined_zrange(wlist: list, _zrange: dict) -> dict:
    zrange_dict = {}

    # first set default
    for well in wlist:
        zrange_dict[well.name] = _zrange["default"]
        logger.info("Default: set %s zrange to %s", well.name, _zrange["default"])

    # well name may be a regular expression
    for wreg, intv in _zrange.items():
        for well in wlist:
            if re.match(wreg + "$", well.name):
                zrange_dict[well.name] = intv
                logger.info(
                    "Override default: set %s zrange to %s, based on regex (%s$)",
                    well.name,
                    intv,
                    wreg,
                )

    return zrange_dict


def _config_default_surf_names(primary):
    """Config default surface name"""
    snames = []
    for ino, _surf in enumerate(primary):
        snames.append("SURF" + str(ino + 1))

    return snames


def update_plotsettings(self, cfg):
    """Update plotsettings and output from the config"""

    # note that 'self' here is a XsectsSettings() instance!

    pcfg = cfg["plotsettings"]

    self.title = cfg["title"]

    self.design_style = pcfg["design"]["style"]
    self.design_zrange = pcfg["design"]["zrange"]
    self.design_gridlines = pcfg["design"]["gridlines"]
    self.design_subtitle = pcfg["design"]["subtitle"]
    self.design_legendsize = pcfg["design"]["legendsize"]
    self.design_legends = pcfg["design"]["legends"]
    self.design_dpi = pcfg["design"]["dpi"]

    self.wellmap_otherwells = pcfg["wellmap"]["otherwells"]
    self.wellmap_expand = pcfg["wellmap"]["expand"]

    self.surf_primary_colors = pcfg["surfaces"]["primary"]["colors"]
    self.surf_primary_fill = pcfg["surfaces"]["primary"]["fill"]

    self.surf_secondary_colors = pcfg["surfaces"]["secondary"]["colors"]
    self.surf_secondary_legend = pcfg["surfaces"]["secondary"]["legend"]

    self.surf_contacts_colors = pcfg["surfaces"]["contacts"]["colors"]

    self.cube_colors = pcfg["cube"]["colors"]
    self.cube_range = pcfg["cube"]["range"]
    self.cube_interpolation = pcfg["cube"]["interpolation"]
    self.cube_alpha = pcfg["cube"]["alpha"]
    self.cube_sampling = pcfg["cube"]["sampling"]

    self.grid_colors = pcfg["grid"]["colors"]
    self.grid_range = pcfg["grid"]["range"]
    self.grid_alpha = pcfg["grid"]["alpha"]
    self.grid_zinc = pcfg["grid"]["zinc"]

    self.wells_zonelog_colors = pcfg["wells"]["zonelog"]["colors"]
    self.wells_zonelog_zoneshift = pcfg["wells"]["zonelog"]["zoneshift"]
    self.wells_zonelog_colordict = pcfg["wells"]["zonelog"]["colordict"]
    self.wells_facieslog_colors = pcfg["wells"]["facieslog"]["colors"]
    self.wells_facieslog_colordict = pcfg["wells"]["facieslog"]["colordict"]
    self.wells_perflog_colors = pcfg["wells"]["perflog"]["colors"]
    self.wells_perflog_colordict = pcfg["wells"]["perflog"]["colordict"]
    self.wells_wellcrossings_show = pcfg["wells"]["wellcrossings"]["show"]
    self.wells_wellcrossings_sampling = pcfg["wells"]["wellcrossings"]["sampling"]
    self.wells_wellcrossings_wfilter = pcfg["wells"]["wellcrossings"]["wfilter"]

    self.output_plotfolder = cfg["output"]["plotfolder"]
    self.output_format = cfg["output"]["format"]
    self.output_prefix = cfg["output"]["prefix"]
    self.output_pdfjoin = cfg["output"]["pdfjoin"]
    self.output_cleanup = cfg["output"]["cleanup"]



import matplotlib.pyplot as plt
import matplotlib.testing as mtst
import numpy as np
import xtgeo
from matplotlib.testing.compare import compare_images

from xtgeoviz.plot.quickplot import quickplot


def test_very_basic_to_file(tmp_path):
    """Just test that matplotlib works, to a file."""
    plt.title("Hello world")
    plt.savefig(tmp_path / "verysimple.png")

    assert (tmp_path / "verysimple.png").is_file() is True


def test_quickplot(tmp_path):
    """Test quickplot of a regular surface."""

    surf = xtgeo.RegularSurface(
        ncol=40, nrow=30, xinc=25, yinc=35, values=np.arange(40 * 30)
    )

    mtst.set_font_settings_for_testing()
    mtst.set_reproducibility_for_testing()
    mtst.set_reproducibility_for_testing()

    quickplot(surf, filename=tmp_path / "quickplot1.png")

    expected = "tests/baseline_images/test_regularsurface/quickplot1.png"
    actual = tmp_path / "quickplot1.png"

    assert compare_images(expected, actual, 50.0) is None



"""Private routines for plotting of xsections."""

from __future__ import annotations

import contextlib
import logging
import os
import os.path
import shutil
import subprocess

import pandas as pd

from xtgeoviz.plot import XSection

logger = logging.getLogger(__name__)


def plotting(self):
    """Given the wells, the surfaces, and the config; make plots."""

    pset = self.plotsettings

    if pset.general_engine == "matplotlib":
        _plotting_mpl(self, pset)


# def _plotting_mpl(config, wells, surfaces, outline, cube):
def _plotting_mpl(self, pset):
    """Plotting of xsections with matplotlib.

    Support for polylines and multiwells will come later. It might be
    that instead of calling xtgeo's XSection, a local hidden xsect class here is
    more practical and flexible.
    """

    _folder_work(pset)

    wellcross = _compute_wellcrossings(pset)

    plotfiles = []

    for well in self.wells["wlist"]:
        logger.info("Plot cross section for well %s", well.name)

        if isinstance(pset.design_zrange, dict):
            zrange_min, zrange_max = pset.design_zrange[well.name]
        else:
            zrange_min, zrange_max = pset.design_zrange

        xplot = XSection(
            zmin=zrange_min,
            zmax=zrange_max,
            well=well,
            surfaces=self.surfaces["primary"],
            zonelogshift=pset.wells_zonelog_zoneshift,
            outline=self.outline,
            colormap=pset.surf_primary_colors,
            cube=self.cube,
            grid=self.grid,
            gridproperty=self.gridproperty,
        )

        xplot.colormap_facies = pset.wells_facieslog_colors
        xplot.colormap_facies_dict = pset.wells_facieslog_colordict

        xplot.colormap_perf = pset.wells_perflog_colors
        xplot.colormap_perf_dict = pset.wells_perflog_colordict

        xplot.colormap_zonelog = pset.wells_zonelog_colors
        xplot.colormap_zonelog_dict = pset.wells_zonelog_colordict

        xplot.legendsize = pset.design_legendsize
        xplot.has_legend = pset.design_legends

        if xplot.fence is None:
            continue

        xplot.canvas(title=well.xwellname, subtitle=pset.design_subtitle)

        if self.cube:
            logger.info("Plot cube backdrop")
            vmin, vmax = (None, None)
            if pset.cube_range:
                vmin, vmax = pset.cube_range

            xplot.plot_cube(
                colormap=pset.cube_colors,
                vmin=vmin,
                vmax=vmax,
                alpha=pset.cube_alpha,
                interpolation=pset.cube_interpolation,
                sampling=pset.cube_sampling,
            )

        if self.grid:
            vmin, vmax = (None, None)
            if pset.grid_range:
                vmin, vmax = pset.grid_range

            xplot.plot_grid3d(
                colormap=pset.grid_colors,
                vmin=vmin,
                vmax=vmax,
                alpha=pset.grid_alpha,
                zinc=pset.grid_zinc,
            )

        logger.info("Plot primary surfaces")
        xplot.plot_surfaces(
            fill=pset.surf_primary_fill,
            axisname="main",
            gridlines=True,
            legend=pset.design_legends,
        )

        wcdf = None
        if wellcross is not None:
            wcdf = wellcross.copy()
            wcdf = wcdf.loc[wcdf["WELL"] == well.xwellname]
            del wcdf["WELL"]

        logger.info("Plot well path")
        xplot.plot_well(
            zonelogname=self.wells["zonelog"],
            facieslogname=self.wells["facieslog"],
            perflogname=self.wells["perflog"],
            wellcrossings=wcdf,
        )

        logger.info("Plot primary again (replot, thin lines)")
        xplot.plot_surfaces(
            fill=False, axisname="lines", legend=False, linewidth=0.3, onecolor="black"
        )

        if self.surfaces["contacts"]:
            xplot.plot_surfaces(
                surfaces=self.surfaces["contacts"],
                legendtitle="Contacts",
                colormap=pset.surf_contacts_colors,
                axisname="contacts",
                legend=pset.design_legends,
            )

        if self.surfaces["secondary"]:
            logger.info("Plot secondary surfaces")
            xplot.plot_surfaces(
                surfaces=self.surfaces["secondary"],
                legendtitle=pset.surf_secondary_legend,
                colormap=pset.surf_secondary_colors,
                axisname="second",
                linewidth=2,
                legend=pset.design_legends,
            )

        if pset.design_legends:
            xplot.plot_map()
            xplot.plot_wellmap(
                expand=pset.wellmap_expand, otherwells=pset.wellmap_otherwells
            )

        pname = _save_fig(well, pset, xplot)
        plotfiles.append(pname)

    _collect_pdf(pset, plotfiles)


def _save_fig(well, pset, xplot):
    prefix = ""
    if pset.output_prefix:
        prefix = pset.output_prefix

    pname = os.path.join(
        pset.output_plotfolder, prefix + well.xwellname + "." + pset.output_format
    )

    xplot.savefig(pname, fformat=pset.output_format, dpi=pset.design_dpi)
    logger.info("Plotting xsection for %s to %s", well.wellname, pname)

    if os.environ.get("XTG_SHOW"):
        xplot.show()

    return pname


def _collect_pdf(pset, plotfiles):
    if pset.output_format != "pdf":
        return

    if pset.output_pdfjoin:
        masterfile = os.path.join([pset.output_plotfolder, pset.output_pdfjoin])

        syscommand = (
            "gs -sDEVICE=pdfwrite -dCompatibilityLevel=1.4 "
            "-dPDFSETTINGS=/default -dNOPAUSE -dQUIET -dBATCH "
            "-dDetectDuplicateImages -dCompressFonts=true -r150 "
            f"-sOutputFile={masterfile} "
        )

        for pfile in plotfiles:
            syscommand = syscommand + pfile + " "

        logger.info("Collecting PDF files to %s", masterfile)
        logger.info(syscommand)
        ier = subprocess.call(syscommand, shell=True)

        if ier != 0:
            raise RuntimeError(f"Received number != 0 from subprosess: {ier}")

        for pfile in plotfiles:
            os.remove(pfile)


def _folder_work(pset):
    if not os.path.exists(pset.output_plotfolder):
        os.makedirs(pset.output_plotfolder)

    if pset.output_cleanup:
        logger.info("Removing previous files in %s", pset.output_plotfolder)
        with contextlib.suppress(Exception):
            shutil.rmtree(pset.output_plotfolder + "/*")


def _compute_wellcrossings(pset):
    """
    Read CSV with wellcrossing or compute well crossing per well and
    collect to a Pandas dataframe

    Note that self.wells_wellcrossing_show can be a pre generated file, or a bool
    but the bool True is not supported yet...
    """
    dfr = None
    if not pset.wells_wellcrossings_show:
        return dfr

    if isinstance(pset.wells_wellcrossings_show, str):
        dfr = pd.read_csv(pset.wells_wellcrossings_show)

    else:
        raise NotImplementedError("Computing wellcrossing built in not supported yet")

    return dfr



# flake8: noqa
from xtgeoviz.frontends import xsectplotting
from xtgeoviz.frontends.xsectplotting import xsectplot
from xtgeoviz.plot.quickplot import quickplot

try:
    from .version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"



import pathlib
from os.path import join

import xtgeo

from xtgeoviz.plot import Map

TPATH = pathlib.Path("../xtgeo-testdata")

SFILE1 = TPATH / "surfaces/reek/1/topreek_rota.gri"
PFILE1 = TPATH / "polygons/reek/1/top_upper_reek_faultpoly.pol"
SFILE2 = TPATH / "surfaces/reek/1/reek_perm_lay1.gri"


def test_simple_plot(tmpdir, generate_plot):
    """Test as simple map plot only making an instance++ and plot."""

    mysurf = xtgeo.surface_from_file(SFILE1)

    # just make the instance, with a lot of defaults behind the scene
    myplot = Map()
    myplot.canvas(title="My o my")
    myplot.colormap = "gist_ncar"
    myplot.plot_surface(mysurf)

    if generate_plot:
        myplot.savefig(join(tmpdir, "map_simple.png"), last=True)
    else:
        myplot.close()


def test_map_plot_with_points(tmpdir, generate_plot):
    """Test as simple map plot with underlying points."""

    mysurf = xtgeo.surface_from_file(SFILE1)

    mypoints = xtgeo.points_from_surface(mysurf)

    df = mypoints.get_dataframe()
    mypoints.set_dataframe(df[::20])

    # just make the instance, with a lot of defaults behind the scene
    myplot = Map()
    myplot.canvas(title="My o my")
    myplot.colormap = "gist_ncar"
    myplot.plot_surface(mysurf)
    myplot.plot_points(mypoints)

    if generate_plot:
        myplot.savefig(join(tmpdir, "map_with_points.png"), last=True)
    else:
        myplot.close()


def test_more_features_plot(tmpdir, generate_plot):
    """Map with some more features added, such as label rotation."""

    mysurf = xtgeo.surface_from_file(SFILE1)

    myfaults = xtgeo.polygons_from_file(PFILE1)

    # just make the instance, with a lot of defaults behind the scene
    myplot = Map()
    myplot.canvas(title="Label rotation")
    myplot.colormap = "rainbow"
    myplot.plot_surface(mysurf, minvalue=1250, maxvalue=2200, xlabelrotation=45)

    myplot.plot_faults(myfaults)

    if generate_plot:
        myplot.savefig(join(tmpdir, "map_more1.png"), last=True)
    else:
        myplot.close()


def test_perm_logarithmic_map(tmpdir, generate_plot):
    """Map with PERM, log scale."""

    mysurf = xtgeo.surface_from_file(SFILE2)

    myplot = Map()
    myplot.canvas(title="PERMX normal scale")
    myplot.colormap = "rainbow"
    myplot.plot_surface(
        mysurf, minvalue=0, maxvalue=6000, xlabelrotation=45, logarithmic=True
    )

    if generate_plot:
        myplot.savefig(join(tmpdir, "permx_normal.png"), last=True)
    else:
        myplot.close()



import glob
import pathlib
from os.path import join

import xtgeo

from xtgeoviz.plot import XSection

TPATH = pathlib.Path("../xtgeo-testdata")

USEFILE1 = TPATH / "wells/reek/1/OP_1.w"
USEFILE2 = TPATH / "surfaces/reek/1/topreek_rota.gri"
USEFILE3 = TPATH / "polygons/reek/1/mypoly.pol"
USEFILE4 = TPATH / "wells/reek/1/OP_5.w"
USEFILE5 = TPATH / "surfaces/reek/2/*.gri"
USEFILE6 = TPATH / "cubes/reek/syntseis_20000101_seismic_depth_stack.segy"

USEFILE7 = TPATH / "wells/reek/1/OP_2.w"


def test_xsection_init():
    """Trigger XSection class, basically."""
    xsect = XSection()
    assert xsect.pagesize == "A4"


def test_simple_plot(tmpdir, show_plot, generate_plot):
    """Test as simple XSECT plot."""

    mywell = xtgeo.well_from_file(USEFILE4)

    mysurfaces = []
    mysurf = xtgeo.surface_from_file(USEFILE2)

    for i in range(10):
        xsurf = mysurf.copy()
        xsurf.values = xsurf.values + i * 20
        xsurf.name = f"Surface_{i}"
        mysurfaces.append(xsurf)

    myplot = XSection(zmin=1500, zmax=1800, well=mywell, surfaces=mysurfaces)

    # set the color table, from file
    clist = [0, 1, 222, 3, 5, 7, 3, 12, 11, 10, 9, 8]
    cfil1 = "xtgeo"
    cfil2 = TPATH / "etc/colortables/colfacies.txt"

    assert 222 in clist
    assert "xtgeo" in cfil1
    assert "colfacies" in str(cfil2)

    myplot.colormap = cfil1

    myplot.canvas(title="Manamana", subtitle="My Dear Well")

    myplot.plot_surfaces(fill=False)

    myplot.plot_well(zonelogname="Zonelog")

    if show_plot:
        myplot.show()

    if generate_plot:
        myplot.savefig(join(tmpdir, "xsect_gbf1.png"), last=True)
    else:
        myplot.close()


def test_simple_plot_with_seismics(tmpdir, show_plot, generate_plot):
    """Test as simple XSECT plot with seismic backdrop."""

    mywell = xtgeo.well_from_file(USEFILE7)
    mycube = xtgeo.cube_from_file(USEFILE6)

    mysurfaces = []
    mysurf = xtgeo.surface_from_file(USEFILE2)

    for i in range(10):
        xsurf = mysurf.copy()
        xsurf.values = xsurf.values + i * 20
        xsurf.name = f"Surface_{i}"
        mysurfaces.append(xsurf)

    myplot = XSection(
        zmin=1000,
        zmax=1900,
        well=mywell,
        surfaces=mysurfaces,
        cube=mycube,
        sampling=10,
        nextend=2,
    )

    # set the color table, from file
    clist = [0, 1, 222, 3, 5, 7, 3, 12, 11, 10, 9, 8]
    cfil1 = "xtgeo"
    cfil2 = TPATH / "etc/colortables/colfacies.txt"

    assert 222 in clist
    assert "xtgeo" in cfil1
    assert "colfacies" in str(cfil2)

    myplot.colormap = cfil1

    myplot.canvas(title="Plot with seismics", subtitle="Some well")

    myplot.plot_cube()
    myplot.plot_surfaces(fill=False)

    myplot.plot_well()

    myplot.plot_map()

    if generate_plot:
        myplot.savefig(join(tmpdir, "xsect_wcube.png"), last=False)

    if show_plot:
        myplot.show()

    myplot.close()


def test_multiple_subplots(tmpdir, show_plot, generate_plot):
    """Test as simple XSECT plot."""

    mywell = xtgeo.well_from_file(USEFILE4)
    mysurf = xtgeo.surface_from_file(USEFILE2)

    myplot = XSection(zmin=1500, zmax=1800, well=mywell, surfaces=[mysurf])

    for i in range(4):
        myplot.canvas(title=str(i), subtitle="My Dear Well")

    myplot.close()


def test_reek1(tmpdir, generate_plot):
    """Test XSect for a Reek well."""

    myfield = xtgeo.polygons_from_file(USEFILE3, fformat="xyz")

    mywells = []
    wnames = glob.glob(str(USEFILE4))
    wnames.sort()
    for wname in wnames:
        mywell = xtgeo.well_from_file(wname)
        mywells.append(mywell)

    mysurfaces = []
    surfnames = glob.glob(str(USEFILE5))
    surfnames.sort()
    for fname in surfnames:
        mysurf = xtgeo.surface_from_file(fname)
        mysurfaces.append(mysurf)

    # Troll lobes
    mylobes = []
    surfnames = glob.glob(str(USEFILE5))
    surfnames.sort()
    for fname in surfnames:
        mysurf = xtgeo.surface_from_file(fname)
        mylobes.append(mysurf)

    for wo in mywells:
        myplot = XSection(
            zmin=1500,
            zmax=1700,
            well=wo,
            surfaces=mysurfaces,
            zonelogshift=-1,
            outline=myfield,
        )

        myplot.canvas(
            title=wo.truewellname,
            subtitle="Before my corrections",
            infotext="Heisan sveisan",
            figscaling=1.2,
        )

        # myplot.colortable('xtgeo')

        myplot.plot_surfaces(fill=True)

        myplot.plot_surfaces(
            surfaces=mylobes,
            fill=False,
            linewidth=4,
            legendtitle="Lobes",
            fancyline=True,
        )

        myplot.plot_well(zonelogname="Zonelog")

        myplot.plot_wellmap()

        myplot.plot_map()

        if generate_plot:
            myplot.savefig(join(tmpdir, "xsect2a.svg"), fformat="svg", last=False)
            myplot.savefig(join(tmpdir, "xsect2a.png"), fformat="png")
        else:
            myplot.close()



"""Module for fast XSection plots of wells/surfaces etc, using matplotlib."""

from __future__ import annotations

import logging
import math
import warnings
from typing import TYPE_CHECKING, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
import pandas as pd
import xtgeo
from matplotlib import collections as mc
from matplotlib.lines import Line2D

from ._libwrapper import matplotlib_colormap, scipy_gaussianfilter
from .baseplot import BasePlot

if TYPE_CHECKING:
    from matplotlib.colors import LinearSegmentedColormap, ListedColormap


logger = logging.getLogger(__name__)


class XSection(BasePlot):
    """Class for plotting a cross-section of a well.

    Args:
        zmin: Upper level of the plot (top Y axis).
        zmax (float): Lower level of the plot (bottom Y axis).
        well: XTGeo well object.
        surfaces: List of XTGeo RegularSurface objects
        surfacenames: List of surface names (str) for legend
        cube: A XTGeo Cube instance
        grid: A XTGeo Grid instance
        gridproperty: A XTGeo GridProperty instance
        colormap: Name of colormap, e.g. 'Set1'. Default is 'xtgeo'
        outline: XTGeo Polygons object

    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        zmin: Optional[float] = 0,
        zmax: Optional[float] = 9999,
        well: Optional[xtgeo.Well] = None,
        surfaces: Optional[list] = None,
        sampling: Optional[int] = 20,
        nextend: Optional[int] = 5,
        colormap: Optional[Union[str, LinearSegmentedColormap, ListedColormap]] = None,
        zonelogshift: Optional[int] = 0,
        surfacenames: Optional[list] = None,
        cube: Optional[xtgeo.Cube] = None,
        grid: Optional[xtgeo.Grid] = None,
        gridproperty: Optional[xtgeo.GridProperty] = None,
        outline: Optional[xtgeo.Polygons] = None,
    ):
        """Init method."""
        super().__init__()

        self._zmin = zmin
        self._zmax = zmax
        self._well = well
        self._nextend = nextend
        self._sampling = sampling
        self._surfaces = surfaces
        self._surfacenames = surfacenames
        self._cube = cube
        self._grid = grid
        self._gridproperty = gridproperty
        self._zonelogshift = zonelogshift
        self._outline = outline

        self._has_axes = True
        self._has_legend = True

        self._pagesize = "A4"
        self._fence = None
        self._legendtitle = "Zones"
        self._legendsize = 5

        self._ax1 = None
        self._ax2 = None
        self._ax3 = None
        self._allfigs = []

        self._colormap_cube = None
        self._colorlegend_cube = False

        self._colormap_grid = None
        self._colorlegend_grid = False

        if colormap is None:
            self._colormap = matplotlib_colormap("viridis")
        else:
            self.define_colormap(colormap)

        self._colormap_facies = self.define_any_colormap("xtgeo")
        self._colormap_facies_dict = {idx: idx for idx in range(100)}

        self._colormap_perf = self.define_any_colormap("xtgeo")
        self._colormap_perf_dict = {idx: idx for idx in range(100)}

        self._colormap_zonelog = None
        self._colormap_zonelog_dict = {idx: idx for idx in range(100)}
        logger.debug("Initialized XSection instance")

    # ==================================================================================
    # Properties
    # Notice additonal props in base class
    # ==================================================================================
    @property
    def pagesize(self):
        """Returns page size."""
        return self._pagesize

    @property
    def legendsize(self):
        """Returns or set the legend size."""
        return self._legendsize

    @legendsize.setter
    def legendsize(self, lsize):
        self._legendsize = lsize

    @property
    def has_legend(self):
        """Returns or set the legends."""
        return self._has_legend

    @has_legend.setter
    def has_legend(self, value):
        if not isinstance(value, bool):
            raise ValueError("Input is not a bool")

        self._has_legend = value

    @property
    def has_axes(self):
        """Returns or set the axes status."""
        return self._has_axes

    @has_axes.setter
    def has_axes(self, value):
        if not isinstance(value, bool):
            raise ValueError("Input is not a bool")

        self._has_axes = value

    @property
    def colormap_facies(self):
        """Set or get the facies colormap."""
        return self._colormap_facies

    @colormap_facies.setter
    def colormap_facies(self, cmap):
        self._colormap_facies = self.define_any_colormap(cmap)

    @property
    def colormap_zonelog(self):
        """Set or get the zonelog colormap."""
        return self._colormap_zonelog

    @colormap_zonelog.setter
    def colormap_zonelog(self, cmap):
        self._colormap_zonelog = self.define_any_colormap(cmap)

    @property
    def colormap_perf(self):
        """Set or get the perforations colormap."""
        return self._colormap_perf

    @colormap_perf.setter
    def colormap_perf(self, cmap):
        self._colormap_perf = self.define_any_colormap(cmap)

    @property
    def colormap_facies_dict(self):
        """Set or get the facies colormap actual dict table."""
        return self._colormap_facies_dict

    @colormap_facies_dict.setter
    def colormap_facies_dict(self, xdict):
        if not isinstance(xdict, dict):
            raise ValueError("Input is not a dict")
        self._colormap_facies_dict = xdict

    @property
    def colormap_perf_dict(self):
        """Set or get the perf colormap actual dict table."""
        return self._colormap_perf_dict

    @colormap_perf_dict.setter
    def colormap_perf_dict(self, xdict):
        if not isinstance(xdict, dict):
            raise ValueError("Input is not a dict")
        self._colormap_perf_dict = xdict

    @property
    def colormap_zonelog_dict(self):
        """Set or get the zonelog colormap actual dict table."""
        return self._colormap_zonelog_dict

    @colormap_zonelog_dict.setter
    def colormap_zonelog_dict(self, xdict):
        if not isinstance(xdict, dict):
            raise ValueError("Input is not a dict")
        self._colormap_zonelog_dict = xdict

    @property
    def fence(self):
        """Set or get the fence spesification."""
        if self._fence is None:
            if self._well is not None:
                wfence = self._well.get_fence_polyline(
                    sampling=self._sampling, nextend=self._nextend, tvdmin=self._zmin
                )
                self._fence = wfence

                if wfence is False:
                    self._fence = None
            else:
                raise ValueError("Input well is None")  # should be more flexible
        return self._fence

    @fence.setter
    def fence(self, myfence):
        # this can be extended with checks and various types of input...
        self._fence = myfence

    # ==================================================================================
    # Functions methods (public)
    # ==================================================================================

    def canvas(self, title=None, subtitle=None, infotext=None, figscaling=1.0):
        """Prepare the canvas to plot on, with title and subtitle.

        Args:
            title (str, optional): Title of plot.
            subtitle (str, optional): Sub title of plot.
            infotext (str, optional): Text to be written as info string.
            figscaling (str, optional): Figure scaling, default is 1.0


        """
        # overriding the base class canvas

        plt.rcParams["axes.xmargin"] = 0  # fill the plot margins

        if not self._has_axes:
            plt.rcParams["axes.titlecolor"] = (0, 0, 0, 0)
            plt.rcParams["axes.edgecolor"] = (0, 0, 0, 0)
            plt.rcParams["axes.labelcolor"] = (0, 0, 0, 0)
            plt.rcParams["axes.titlecolor"] = (0, 0, 0, 0)
            plt.rcParams["xtick.color"] = (0, 0, 0, 0)
            plt.rcParams["ytick.color"] = (0, 0, 0, 0)

        self._fig = plt.figure(figsize=(11.69 * figscaling, 8.27 * figscaling))
        self._allfigs.append(self._fig)

        ax1 = {
            "main": plt.subplot2grid((20, 28), (0, 0), rowspan=20, colspan=23),
        }

        ax2 = plt.subplot2grid(
            (20, 28), (10, 23), rowspan=5, colspan=5, frame_on=self._has_legend
        )

        ax3 = plt.subplot2grid(
            (20, 28), (15, 23), rowspan=5, colspan=5, frame_on=self._has_legend
        )

        if self._has_legend:
            # indicate A to B
            plt.text(
                0.02,
                0.98,
                "A",
                ha="left",
                va="top",
                transform=ax1["main"].transAxes,
                fontsize=8,
            )
            plt.text(
                0.98,
                0.98,
                "B",
                ha="right",
                va="top",
                transform=ax1["main"].transAxes,
                fontsize=8,
            )

        # title here:
        if title is not None:
            plt.text(
                0.5,
                1.09,
                title,
                ha="center",
                va="center",
                transform=ax1["main"].transAxes,
                fontsize=18,
            )

        if subtitle is not None:
            ax1["main"].set_title(subtitle, size=14)

        if infotext is not None:
            plt.text(
                -0.11,
                -0.11,
                infotext,
                ha="left",
                va="center",
                transform=ax1["main"].transAxes,
                fontsize=6,
            )

        ax1["main"].set_ylabel("Depth", fontsize=12.0)
        ax1["main"].set_xlabel("Length along well", fontsize=12)

        ax2.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            right=False,
            left=False,
            labelbottom=False,
            labeltop=False,
            labelright=False,
            labelleft=False,
        )

        ax3.tick_params(
            axis="both",
            which="both",
            bottom=False,
            top=False,
            right=False,
            left=False,
            labelbottom=False,
            labeltop=False,
            labelright=False,
            labelleft=False,
        )

        # need these also, a bug in functions above?
        ax2.xaxis.set_major_formatter(plt.NullFormatter())
        ax2.yaxis.set_major_formatter(plt.NullFormatter())
        ax3.xaxis.set_major_formatter(plt.NullFormatter())
        ax3.yaxis.set_major_formatter(plt.NullFormatter())

        self._ax1 = ax1
        self._ax2 = ax2
        self._ax3 = ax3

    def plot_well(
        self,
        zonelogname="ZONELOG",
        facieslogname=None,
        perflogname=None,
        wellcrossings=None,
        wellcrossingnames=True,
        wellcrossingyears=False,
        welltrajcolor="black",
        welltrajwidth=6,
    ):
        """Input an XTGeo Well object and plot it."""
        if self.fence is None:
            return

        # reduce the well data by Pandas operations
        dfr = self._well.get_dataframe(copy=False)
        dfr = dfr[dfr["Z_TVDSS"] > self._zmin]

        # Create a relative XYLENGTH vector (0.0 where well starts)
        self._well.create_relative_hlen()

        dfr = self._well.get_dataframe(copy=False)
        if dfr.empty:
            self._showok = False
            return

        # get the well trajectory (numpies) as copy
        zv = dfr["Z_TVDSS"].values.copy()
        hv = dfr["R_HLEN"].values.copy()

        # plot the perflog, if any, first
        if perflogname:
            ax, bba = self._currentax(axisname="perf")
            self._plot_well_perflog(dfr, ax, bba, perflogname, legend=self._has_legend)

        # plot the facies, if any, behind the trajectory; ie. first or second
        if facieslogname:
            ax, bba = self._currentax(axisname="facies")
            self._plot_well_faclog(dfr, ax, bba, facieslogname, legend=self._has_legend)

        axx, _bbxa = self._currentax(axisname="well")
        self._plot_well_traj(
            axx, zv, hv, welltrajcolor=welltrajcolor, linewidth=welltrajwidth
        )

        if zonelogname:
            ax, bba = self._currentax(axisname="main")
            self._plot_well_zlog(dfr, axx, bba, zonelogname, legend=self._has_legend)

        if wellcrossings is not None and wellcrossings.empty:
            wellcrossings = None

        if wellcrossings is not None:
            self._plot_well_crossings(
                dfr, axx, wellcrossings, wellcrossingnames, wellcrossingyears
            )

    def set_xaxis_md(self, gridlines=False):
        """Set x-axis labels to measured depth."""
        md_start = self._well.get_dataframe(copy=False)["MDEPTH"].iloc[0]
        md_start_round = int(math.floor(md_start / 100.0)) * 100
        md_start_delta = md_start - md_start_round

        auto_ticks = plt.xticks()
        auto_ticks_delta = auto_ticks[0][1] - auto_ticks[0][0]

        ax, _ = self._currentax(axisname="main")
        lim = ax.get_xlim()

        new_ticks = []
        new_tick_labels = []
        delta = 0
        for tick in auto_ticks[0]:
            new_ticks.append(int(float(tick) - md_start_delta))
            new_tick_labels.append(int(md_start_round + delta))
            delta += auto_ticks_delta

        # Set new xticks and labels
        plt.xticks(new_ticks, new_tick_labels)

        if gridlines:
            ax.tick_params(axis="y", direction="in", which="both")
            ax.minorticks_on()
            ax.grid(color="black", linewidth=0.8, which="major", linestyle="-")
            ax.grid(color="black", linewidth=0.5, which="minor", linestyle="--")

        # Restore xaxis limits and set axis title
        ax.set_xlim(lim)
        ax.set_ylabel("TVD MSL [m]", fontsize=12.0)
        ax.set_xlabel("Measured Depth [m]", fontsize=12)

    def _plot_well_traj(self, ax, zv, hv, welltrajcolor, linewidth):
        """Plot the trajectory as a black line."""
        zv_copy = ma.masked_where(zv < self._zmin, zv)
        hv_copy = ma.masked_where(zv < self._zmin, hv)

        ax.plot(hv_copy, zv_copy, linewidth=linewidth, c=welltrajcolor)

    @staticmethod
    def _line_segments_colors(df, idx, ctable, logname, fillnavalue):
        """Get segment and color array for plotting matplotlib lineCollection."""
        df_idx = pd.DataFrame(
            {"idx_log": list(idx.keys()), "idx_color": list(idx.values())}
        )

        df_ctable = df_idx.merge(
            pd.DataFrame({"ctable": ctable}),
            how="left",
            left_on="idx_color",
            right_index=True,
        )

        dff = df.merge(df_ctable, how="left", left_on=logname, right_on="idx_log")

        dff["point"] = list(zip(dff["R_HLEN"], dff["Z_TVDSS"]))

        # find line segments
        segments = []
        segments_i = -1
        colorlist = []
        previous_color = None

        for point, color in zip(dff["point"], dff["ctable"]):
            if np.any(np.isnan(color)):
                color = fillnavalue

            if color == previous_color:
                segments[segments_i].append(point)
                previous_color = color
            else:
                # add endpoint to current segment
                if segments_i > 0:
                    segments[segments_i].append(point)

                # start new segment
                segments.append([point])
                colorlist.append(color)

                previous_color = color
                segments_i += 1

        colorlist = np.asarray(colorlist, dtype=object)

        return segments, colorlist

    def _plot_well_zlog(self, df, ax, bba, zonelogname, logwidth=4, legend=False):
        """Plot the zone log as colored segments."""
        if zonelogname not in df.columns:
            return

        zomin = 0
        zomax = 0

        try:
            zomin = int(df[zonelogname].min())
            zomax = int(df[zonelogname].max())
        except ValueError:
            self._showok = False
            return

        logger.debug("ZOMIN, ZOMAX %s %s ?", zomin, zomax)  # used for what?

        zshift = 0
        if self._zonelogshift != 0:
            zshift = self._zonelogshift

        if self.colormap_zonelog is not None:
            cmap = self.colormap_zonelog
            ctable = self.get_any_colormap_as_table(cmap)
        else:
            ctable = self.get_colormap_as_table()

        idx = self.colormap_zonelog_dict

        # adjust for zoneshift.
        idx_zshift = {}
        for key in idx:
            idx_zshift[key - zshift + 1] = idx[key]

        fillnavalue = (0.9, 0.9, 0.9)
        segments, segments_colors = self._line_segments_colors(
            df, idx_zshift, ctable, zonelogname, fillnavalue
        )

        lc = mc.LineCollection(
            segments, colors=segments_colors, linewidth=logwidth, zorder=202
        )

        ax.add_collection(lc)

        if legend:
            zrecord = self._well.get_logrecord(zonelogname)
            zrecord = {val: zname for val, zname in zrecord.items() if val >= 0}

            zcolors = {}
            for zone in zrecord:
                zcolors[zrecord[zone]] = (
                    idx[zone] if isinstance(idx[zone], str) else ctable[idx[zone]]
                )

            self._drawproxylegend(ax, bba, items=zcolors, title="Zonelog")

    def _plot_well_faclog(self, df, ax, bba, facieslogname, logwidth=9, legend=True):
        """Plot the facies log as colored segments.

        Args:
            df (dataframe): The Well dataframe.
            ax (axes): The ax plot object.
            bba: Bounding box
            facieslogname (str): name of the facies log.
            logwidth (int): Log linewidth.
            legend (bool): Plot log legend?
        """
        if facieslogname not in df.columns:
            return

        cmap = self.colormap_facies
        ctable = self.get_any_colormap_as_table(cmap)
        idx = self.colormap_facies_dict

        fillnavalue = (0, 0, 0, 0)  # transparent
        segments, segments_colors = self._line_segments_colors(
            df, idx, ctable, facieslogname, fillnavalue
        )

        lc = mc.LineCollection(
            segments, colors=segments_colors, linewidth=logwidth, zorder=201
        )

        ax.add_collection(lc)

        if legend:
            frecord = self._well.get_logrecord(facieslogname)
            frecord = {val: fname for val, fname in frecord.items() if val >= 0}

            fcolors = {}
            for facies in frecord:
                if isinstance(idx[facies], str):
                    color = idx[facies]
                else:
                    color = ctable[idx[facies]]

                fcolors[frecord[facies]] = color

            self._drawproxylegend(ax, bba, items=fcolors, title="Facies")

    def _plot_well_perflog(self, df, ax, bba, perflogname, logwidth=12, legend=True):
        """Plot the perforation log as colored segments.

        Args:
            df (dataframe): The Well dataframe.
            ax (axes): The ax plot object.
            zv (ndarray): The numpy Z TVD array.
            bba: Boundinng box
            hv (ndarray): The numpy Length  array.
            perflogname (str): name of the perforation log.
            logwidth (int): Log linewidth.
            legend (bool): Plot log legend?
        """
        if perflogname not in df.columns:
            return

        cmap = self.colormap_perf
        ctable = self.get_any_colormap_as_table(cmap)
        idx = self.colormap_perf_dict

        fillnavalue = (0, 0, 0, 0)  # transparent
        segments, segments_colors = self._line_segments_colors(
            df, idx, ctable, perflogname, fillnavalue
        )

        lc = mc.LineCollection(
            segments, colors=segments_colors, linewidth=logwidth, zorder=200
        )

        ax.add_collection(lc)

        if legend:
            precord = self._well.get_logrecord(perflogname)
            precord = {val: pname for val, pname in precord.items() if val >= 0}

            pcolors = {}
            for perf in precord:
                pcolors[precord[perf]] = (
                    idx[perf] if isinstance(idx[perf], str) else ctable[idx[perf]]
                )

            self._drawproxylegend(ax, bba, items=pcolors, title="Perforations")

    @staticmethod
    def _plot_well_crossings(dfr, ax, wcross, names=True, years=False):
        """Plot well crossing based on dataframe (wcross).

        The well crossing coordinates are identified for this well,
        and then it is looking for the closest coordinate. Given this
        coordinate, a position is chosen.

        The pandas dataframe wcross shall have the following columns:

        * Name of crossing wells named CWELL
        * Coordinate X named X_UTME
        * Coordinate Y named Y_UTMN
        * Coordinate Z named Z_TVDSS

        Optional column:
        * Drilled year of crossing well named CYEAR

        Args:
            dfr: Well dataframe
            ax: current axis
            wcross: A pandas dataframe with precomputed well crossings
            names: Display the names of the crossed wells
            years: Display the drilled year of the crossed wells
        """
        placings = {
            0: (40, 40),
            1: (40, -20),
            2: (-30, 30),
            3: (30, 20),
            4: (-40, 30),
            5: (-20, 40),
        }

        for index, row in wcross.iterrows():
            xcoord = row.X_UTME
            ycoord = row.Y_UTMN

            dfrc = dfr.copy()

            dfrc["DLEN"] = pow(
                pow(dfrc.X_UTME - xcoord, 2) + pow(dfrc.Y_UTMN - ycoord, 2), 0.5
            )

            minindx = dfrc.DLEN.idxmin()

            ax.scatter(
                dfrc.R_HLEN[minindx],
                row.Z_TVDSS,
                marker="o",
                color="black",
                s=70,
                zorder=300,
            )
            ax.scatter(
                dfrc.R_HLEN[minindx],
                row.Z_TVDSS,
                marker="o",
                color="orange",
                s=38,
                zorder=302,
            )

            modulo = index % 5

            text = ""
            if names:
                text = xtgeo.Well.get_short_wellname(row.CWELL)

            if years:
                text = text + "\n" + row.CYEAR if names else row.CYEAR

            if names or years:
                ax.annotate(
                    text,
                    size=6,
                    xy=(dfrc.R_HLEN[minindx], row.Z_TVDSS),
                    xytext=placings[modulo],
                    textcoords="offset points",
                    arrowprops={
                        "arrowstyle": "->",
                        "connectionstyle": "angle3,angleA=0,angleB=90",
                    },
                    color="black",
                )

    def _drawproxylegend(self, ax, bba, items, title=None):
        proxies = []
        labels = []

        for item in items:
            color = items[item]
            proxies.append(Line2D([0, 1], [0, 1], color=color, linewidth=5))
            labels.append(item)

        ax.legend(
            proxies,
            labels,
            loc="upper left",
            bbox_to_anchor=bba,
            prop={"size": self._legendsize},
            title=title,
            handlelength=2,
        )

    def _drawlegend(self, ax, bba, title=None):
        leg = ax.legend(
            loc="upper left",
            bbox_to_anchor=bba,
            prop={"size": self._legendsize},
            title=title,
            handlelength=2,
        )

        for myleg in leg.get_lines():
            myleg.set_linewidth(5)

    def _currentax(self, axisname="main"):
        """Keep track of current axis; is needed as one new legend need one new axis."""
        # for multiple legends, bba is dynamic
        bbapos = {
            "main": (1.22, 1.12, 1, 0),
            "contacts": (1.01, 1.12),
            "second": (1.22, 0.50),
            "facies": (1.01, 1.00),
            "perf": (1.22, 0.45),
        }

        ax1 = self._ax1

        if axisname != "main":
            ax1[axisname] = self._ax1["main"].twinx()

            # invert min,max to invert the Y axis
            ax1[axisname].set_ylim([self._zmax, self._zmin])

            ax1[axisname].set_yticklabels([])
            ax1[axisname].tick_params(axis="y", direction="in")

        ax = self._ax1[axisname]

        bba = bbapos.get(axisname, (1.22, 0.5))

        return ax, bba

    def plot_cube(
        self,
        colormap="seismic",
        vmin=None,
        vmax=None,
        alpha=0.7,
        interpolation="gaussian",
        sampling="nearest",
    ):
        """Plot a cube backdrop.

        Args:
            colormap (ColorMap): Name of color map (default 'seismic')
            vmin (float): Minimum value in plot.
            vmax (float); Maximum value in plot
            alpha (float): Alpah blending number beween 0 and 1.
            interpolation (str): Interpolation for plotting, cf. matplotlib
                documentation on this. Also gaussianN is allowed, where
                N = 1..9.
            sampling (str): 'nearest' (default) or 'trilinear' (more precise)

        Raises:
            ValueError: No cube is loaded

        """
        if self.fence is None:
            return

        if self._cube is None:
            raise ValueError("Ask for plot cube, but noe cube is loaded")

        ax, _ = self._currentax(axisname="main")

        zinc = self._cube.zinc / 2.0

        zvv = self._cube.get_randomline(
            self.fence,
            zmin=self._zmin,
            zmax=self._zmax,
            zincrement=zinc,
            sampling=sampling,
        )

        h1, h2, v1, v2, arr = zvv

        if self._colormap_cube is None:
            if colormap is None:
                colormap = "seismic"
            self._colormap_cube = self.define_any_colormap(colormap)

        if "gaussian" in interpolation:  # allow gaussian3 etc
            nnv = interpolation[-1]
            try:
                nnv = int(nnv)
                arr = scipy_gaussianfilter(arr, nnv)
                interpolation = "none"
            except ValueError:
                interpolation = "gaussian"

        img = ax.imshow(
            arr,
            cmap=self._colormap_cube,
            interpolation=interpolation,
            vmin=vmin,
            vmax=vmax,
            extent=(h1, h2, v2, v1),
            aspect="auto",
            alpha=alpha,
        )

        # steer this?
        if self._colorlegend_cube:
            self._fig.colorbar(img, ax=ax)

    def plot_grid3d(
        self,
        colormap="rainbow",
        vmin=None,
        vmax=None,
        alpha=0.7,
        zinc=0.5,
        interpolation="auto",
    ):
        """Plot a sampled grid with gridproperty backdrop.

        Args:
            colormap (ColorMap): Name of color map (default 'rainbow')
            vmin (float): Minimum value in plot.
            vmax (float); Maximum value in plot
            alpha (float): Alpha blending number beween 0 and 1.
            zinc (float): Sampling vertically, default is 0.5
            interpolation (str): Interpolation for plotting, cf. matplotlib
                documentation on this. "auto" uses "nearest" for discrete
                parameters and "antialiased" for floats.

        Raises:
            ValueError: No grid or gridproperty is loaded

        """
        if self.fence is None:
            return

        if self._grid is None or self._gridproperty is None:
            raise ValueError("Ask for plot of grid, but no grid is loaded")

        ax, _bba = self._currentax(axisname="main")

        zvv = self._grid.get_randomline(
            self.fence,
            self._gridproperty,
            zmin=self._zmin,
            zmax=self._zmax,
            zincrement=zinc,
        )

        h1, h2, v1, v2, arr = zvv

        if self._colormap_grid is None:
            if colormap is None:
                colormap = "rainbow"
            self._colormap_grid = self.define_any_colormap(colormap)

        if interpolation == "auto":
            if self._gridproperty.isdiscrete:
                interpolation = "nearest"
            else:
                interpolation = "antialiased"

        img = ax.imshow(
            arr,
            cmap=self._colormap_grid,
            vmin=vmin,
            vmax=vmax,
            extent=(h1, h2, v2, v1),
            aspect="auto",
            alpha=alpha,
            interpolation=interpolation,
        )

        # steer this?
        if self._colorlegend_grid:
            self._fig.colorbar(img, ax=ax)

    def plot_surfaces(
        self,
        fill=False,
        surfaces=None,
        surfacenames=None,
        colormap=None,
        onecolor=None,
        linewidth=1.0,
        linestyle="-",
        legend=True,
        legendtitle=None,
        fancyline=False,
        axisname="main",
        gridlines=False,
    ):  # pylint: disable=too-many-branches, too-many-statements
        """Input a surface list (ordered from top to base) , and plot them."""
        if self.fence is None:
            return

        ax, bba = self._currentax(axisname=axisname)

        # either use surfaces from __init__, or override with surfaces
        # speciefied here
        if surfaces is None:
            surfaces = self._surfaces
            surfacenames = self._surfacenames

        surfacenames = [surf.name for surf in surfaces]

        if legendtitle is None:
            legendtitle = self._legendtitle

        if colormap is None:
            colormap = self._colormap
        else:
            self.define_colormap(colormap)

        nlen = len(surfaces)

        # legend
        slegend = []
        if surfacenames is None:
            for i in range(nlen):
                slegend.append(f"Surf {i}")

        else:
            # do a check
            if len(surfacenames) != nlen:
                msg = (
                    "Wrong number of entries in surfacenames! "
                    "Number of names is {} while number of files "
                    "is {}".format(len(surfacenames), nlen)
                )
                raise ValueError(msg)

            slegend = surfacenames

        if nlen > self._colormap.N:
            raise ValueError(
                f"Too few colors in color table ({self._colormap.N}) "
                f"vs number of surfaces ({nlen})"
            )

        # sample the horizon to the fence:
        colortable = self.get_colormap_as_table()
        for i in range(nlen):
            usecolor = colortable[i]
            if onecolor:
                usecolor = onecolor
            if not fill:
                hfence = surfaces[i].get_randomline(self.fence)
                xcol = "white"
                if fancyline:
                    cxx = usecolor
                    if cxx[0] + cxx[1] + cxx[2] > 1.5:
                        xcol = "black"
                    ax.plot(
                        hfence[:, 0], hfence[:, 1], linewidth=1.2 * linewidth, c=xcol
                    )
                ax.plot(
                    hfence[:, 0],
                    hfence[:, 1],
                    linewidth=linewidth,
                    c=usecolor,
                    label=slegend[i],
                    linestyle=linestyle,
                )
                if fancyline:
                    ax.plot(
                        hfence[:, 0], hfence[:, 1], linewidth=0.3 * linewidth, c=xcol
                    )
            else:
                # need copy() .. why?? found by debugging...
                hfence1 = surfaces[i].get_randomline(self.fence).copy()
                x1 = hfence1[:, 0]
                y1 = hfence1[:, 1]
                if i < (nlen - 1):
                    hfence2 = surfaces[i + 1].get_randomline(self.fence).copy()
                    y2 = hfence2[:, 1]
                else:
                    y2 = y1.copy()

                ax.plot(
                    x1, y1, linewidth=0.1 * linewidth, linestyle=linestyle, c="black"
                )
                ax.fill_between(x1, y1, y2, facecolor=colortable[i], label=slegend[i])

        # invert min,max to invert the Y axis
        ax.set_ylim([self._zmax, self._zmin])

        if legend:
            self._drawlegend(ax, bba, title=legendtitle)

        if axisname != "main":
            ax.set_yticklabels([])

        ax.tick_params(axis="y", direction="in")

        if axisname == "main" and gridlines:
            ax.grid(color="grey", linewidth=0.2)

    def plot_md_data(
        self,
        data=None,
        markersize=10,
        color="red",
        linestyle="",
        label=False,
        zorder=350,
        **kwargs,
    ):
        """Plot MD vs TVD data as lines and/or markers.

        The input pandas dataframe points shall have the following columns:
        * Name of well(s) named WELL
        * Coordinate X named MDEPTH
        * Coordinate Y named Z_TVDSS
        """
        ax, _ = self._currentax(axisname="main")

        well = self._well
        data_well = data.copy()
        data_well = data_well.loc[data_well["WELL"] == well.xwellname]
        del data_well["WELL"]

        md_start = well.get_dataframe(copy=False)["MDEPTH"].iloc[0]
        data_well["R_HLEN"] = data_well["MDEPTH"]
        data_well["R_HLEN"] = data_well["R_HLEN"].subtract(md_start)

        data_well.plot(
            ax=ax,
            x="R_HLEN",
            y="Z_TVDSS",
            legend=None,
            linestyle=linestyle,
            markersize=markersize,
            color=color,
            label=label,
            zorder=zorder,
            **kwargs,
        )

    def plot_wellmap(self, otherwells=None, expand=1):
        """Plot well map as local view, optionally with nearby wells.

        Args:
            otherwells (list of Polygons): List of surrounding wells to plot,
                these wells are repr as Polygons instances, one per well.
            expand (float): Plot axis expand factor (default is 1); larger
                values may be used if other wells are plotted.


        """
        ax = self._ax2

        if self.fence is not None:
            xwellarray = self._well.get_dataframe(copy=False)["X_UTME"].values
            ywellarray = self._well.get_dataframe(copy=False)["Y_UTMN"].values

            ax.plot(xwellarray, ywellarray, linewidth=4, c="cyan")

            ax.plot(self.fence[:, 0], self.fence[:, 1], linewidth=1, c="black")
            ax.annotate("A", xy=(self.fence[0, 0], self.fence[0, 1]), fontsize=8)
            ax.annotate("B", xy=(self.fence[-1, 0], self.fence[-1, 1]), fontsize=8)
            ax.set_aspect("equal", "datalim")

            left, right = ax.get_xlim()
            xdiff = right - left
            bottom, top = ax.get_ylim()
            ydiff = top - bottom

            ax.set_xlim(left - (expand - 1.0) * xdiff, right + (expand - 1.0) * xdiff)
            ax.set_ylim(bottom - (expand - 1.0) * ydiff, top + (expand - 1.0) * ydiff)
        if otherwells:
            for poly in otherwells:
                if not isinstance(poly, xtgeo.Polygons):
                    warnings.warn(
                        f"<otherw> not a Polygons instance, but a {type(poly)}",
                        UserWarning,
                    )
                    continue
                if poly.name == self._well.xwellname:
                    continue
                xwp = poly.get_dataframe(copy=False)[poly.xname].values
                ywp = poly.get_dataframe(copy=False)[poly.yname].values
                ax.plot(xwp, ywp, linewidth=1, c="grey")
                ax.annotate(poly.name, xy=(xwp[-1], ywp[-1]), color="grey", size=5)

    def plot_map(self):
        """Plot well location map as an overall view (with field outline)."""
        if self.fence is None:
            return

        if not self._outline:
            return

        ax = self._ax3
        if self.fence is not None:
            xp = self._outline.get_dataframe(copy=False)["X_UTME"].values
            yp = self._outline.get_dataframe(copy=False)["Y_UTMN"].values
            ip = self._outline.get_dataframe(copy=False)["POLY_ID"].values

            ax.plot(self._fence[:, 0], self._fence[:, 1], linewidth=3, c="red")

            for i in range(int(ip.min()), int(ip.max()) + 1):
                xpc = xp.copy()[ip == i]
                ypc = yp.copy()[ip == i]
                if len(xpc) > 1:
                    ax.plot(xpc, ypc, linewidth=0.3, c="black")

            ax.set_aspect("equal", "datalim")



"""The baseplot module."""

from __future__ import annotations

import logging
import warnings

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap

from . import _colortables as _ctable
from ._libwrapper import matplotlib_colormap

logger = logging.getLogger(__name__)


class BasePlot:
    """Base class for plots, proviridisding some functions to share."""

    def __init__(self):
        """Init method."""

        self._contourlevels = 3

        # wants self._color to be LinearSegmentedColormap instance
        viridis = matplotlib_colormap("viridis")
        if isinstance(viridis, ListedColormap):
            self._colormap = LinearSegmentedColormap.from_list(
                "viridis", viridis.colors, N=viridis.N
            )
        elif isinstance(viridis, LinearSegmentedColormap):
            self._colormap = viridis
        else:
            raise ValueError("Cannot initialize color attribute")

        self._ax = None
        self._tight = False
        self._showok = True
        self._fig = None
        self._allfigs = []
        self._pagesize = "A4"

        logger.debug("Ran __init__ for BasePlot")

    @property
    def contourlevels(self):
        """Get or set the number of contour levels."""
        return self._contourlevels

    @contourlevels.setter
    def contourlevels(self, num):
        self._contourlevels = num

    @property
    def colormap(self):
        """Get or set the color table as a matplot cmap object."""
        return self._colormap

    @colormap.setter
    def colormap(self, cmap):
        if isinstance(cmap, LinearSegmentedColormap):
            self._colormap = cmap
        elif isinstance(cmap, ListedColormap):
            self._colormap = LinearSegmentedColormap.from_list(
                cmap.name, cmap.colors, N=cmap.N
            )
        elif isinstance(cmap, str):
            logger.debug("Definition of a colormap from string name: %s", cmap)
            self.define_colormap(cmap)
        else:
            raise ValueError("Input incorrect")

        logger.debug("Colormap: %s", self._colormap)

    @property
    def pagesize(self):
        """Returns page size."""
        return self._pagesize

    @staticmethod
    def define_any_colormap(cfile, colorlist=None):
        """Defines any color map from file or a predefined name.

        This is a static method, which returns a matplotlib CM object.

        Args:
            cfile (str): File name (RMS format) or an alias for a predefined
                map name, e.g. 'xtgeo', or one of matplotlibs numerous tables.
            colorlist (list, int, optional): List of integers redefining
                color entries per zone and/or well, which starts
                from 0 index. Default is just keep the linear sequence as is.

        """
        valid_maps = sorted(m for m in plt.cm.datad)

        logger.debug("Valid color maps: %s", valid_maps)

        colors = []

        cmap = plt.get_cmap("rainbow")

        if cfile is None:
            cfile = "rainbow"
            cmap = plt.get_cmap("rainbow")

        elif cfile == "xtgeo":
            colors = _ctable.xtgeocolors()
            cmap = LinearSegmentedColormap.from_list(cfile, colors, N=len(colors))
            cmap.name = "xtgeo"
        elif cfile == "random40":
            colors = _ctable.random40()
            cmap = LinearSegmentedColormap.from_list(cfile, colors, N=len(colors))
            cmap.name = "random40"

        elif cfile == "randomc":
            colors = _ctable.randomc(256)
            cmap = LinearSegmentedColormap.from_list(cfile, colors, N=len(colors))
            cmap.name = "randomc"

        elif isinstance(cfile, str) and "rms" in cfile:
            colors = _ctable.colorsfromfile(cfile)
            cmap = LinearSegmentedColormap.from_list("rms", colors, N=len(colors))
            cmap.name = cfile
        elif cfile in valid_maps:
            cmap = plt.get_cmap(cfile)
            for i in range(cmap.N):
                colors.append(cmap(i))
        else:
            warnings.warn(
                "Trying to access as color map not installed in "
                f"this version of matplotlib: <{cfile}>. Revert to <rainbow>",
                UserWarning,
                stacklevel=2,
            )
            cmap = plt.get_cmap("rainbow")
            for i in range(cmap.N):
                colors.append(cmap(i))

        ctable = []

        if colorlist:
            for entry in colorlist:
                if entry < len(colors):
                    ctable.append(colors[entry])
                else:
                    logger.warning("Color list out of range")
                    ctable.append(colors[0])

            cmap = LinearSegmentedColormap.from_list(ctable, colors, N=len(colors))
            cmap.name = "user"

        return cmap

    @staticmethod
    def get_any_colormap_as_table(cmap):
        """Returns the given color map cmap as a list of RGB tuples."""
        return [cmap(i) for i in range(cmap.N)]

    def get_colormap_as_table(self):
        """Get the current color map as a list of RGB tuples."""
        return self.get_any_colormap_as_table(self._colormap)

    def define_colormap(self, cfile, colorlist=None):
        """Defines a color map from file or a predefined name.

        Args:
            cfile (str): File name (RMS format) or an alias for a predefined
                map name, e.g. 'xtgeo', or one of matplotlibs numerous tables.
            colorlist (list, int, optional): List of integers redefining
                color entries per zone and/or well, which starts
                from 0 index. Default is just keep the linear sequence as is.

        """
        logger.debug("Defining colormap")

        cmap = self.define_any_colormap(cfile, colorlist=colorlist)

        self.contourlevels = cmap.N
        self._colormap = cmap

    def canvas(self, title=None, subtitle=None, infotext=None, figscaling=1.0):
        """Prepare the canvas to plot on, with title and subtitle.

        Args:
            title (str, optional): Title of plot.
            subtitle (str, optional): Sub title of plot.
            infotext (str, optional): Text to be written as info string.
            figscaling (str, optional): Figure scaling, default is 1.0


        """
        self._fig, self._ax = plt.subplots(
            figsize=(11.69 * figscaling, 8.27 * figscaling)
        )
        self._allfigs.append(self._fig)
        if title is not None:
            self._fig.suptitle(title, fontsize=18)
        if subtitle is not None:
            self._ax.set_title(subtitle, size=14)
        if infotext is not None:
            self._fig.text(0.01, 0.02, infotext, ha="left", va="center", fontsize=8)

    def show(self):
        """Call to matplotlib.pyplot show method.

        Returns:
            True of plotting is done; otherwise False
        """
        if self._tight:
            self._fig.tight_layout()

        if self._showok:
            logger.debug("Calling plt show method...")
            plt.show()
            return True

        logger.warning("Nothing to plot (well outside Z range?)")
        return False

    def close(self):
        """
        Explicitly closes the plot, meaning that memory will be cleared.
        After close is called, no more operations can be performed on the plot.
        """
        for fig in self._allfigs:
            plt.close(fig)

    def savefig(self, filename, fformat="png", last=True, **kwargs):
        """Call to matplotlib.pyplot savefig method.

        Args:
            filename (str): File to plot to
            fformat (str): Plot format, e.g. png (default), jpg, svg
            last (bool): Default is true, calls close on the plot, let last
                be False for all except the last plots.
            kwargs: Additional keyword arguments that are passed
                to matplotlib when saving the figure

        Returns:
            True of plotting is done; otherwise False

        Example::

            myplot.savefig('TMP/layerslice.svg', fformat='svg', last=False)
            myplot.savefig('TMP/layerslice.png')

        .. versionchanged:: 2.4 added kwargs option

        """
        if self._tight:
            self._fig.tight_layout()

        if self._showok:
            plt.savefig(filename, format=fformat, **kwargs)
            if last:
                self.close()
            return True

        logger.warning("Nothing to plot (well outside Z range?)")
        return False



"""Module for map plots of surfaces, using matplotlib."""

from __future__ import annotations

import logging

import matplotlib.patches as mplp
import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma
from matplotlib import ticker

from .baseplot import BasePlot

logger = logging.getLogger(__name__)


class Map(BasePlot):
    """Class for plotting a map, using matplotlib."""

    def __init__(self):
        """The constructor method for a Map object."""
        super().__init__()

        self._wells = None
        self._surface = None
        self._tight = False

        self._wfence = None
        self._showok = True  # to indicate if plot is OK to show
        self._legendtitle = "Map"

    # =========================================================================
    # Properties
    # =========================================================================
    @property
    def pagesize(self):
        """Returns page size."""
        return self._pagesize

    # =========================================================================
    # Functions methods (public)
    # =========================================================================

    def plot_surface(
        self,
        surf,
        minvalue=None,
        maxvalue=None,
        contourlevels=None,
        xlabelrotation=None,
        colormap=None,
        logarithmic=False,
    ):  # pylint: disable=too-many-statements
        """Input a surface and plot it."""
        # need a deep copy to avoid changes in the original surf

        logger.debug("The key contourlevels %s is not in use", contourlevels)

        usesurf = surf.copy()
        if usesurf.yflip < 0:
            usesurf.swapaxes()

        if abs(surf.rotation) > 0.001:
            usesurf.unrotate()

        xi, yi, zi = usesurf.get_xyz_values()

        zimask = ma.getmaskarray(zi).copy()  # yes need a copy!

        legendticks = None
        if minvalue is not None and maxvalue is not None:
            minv = float(minvalue)
            maxv = float(maxvalue)

            step = (maxv - minv) / 10.0
            legendticks = []
            for i in range(10 + 1):
                llabel = float(f"{minv + step * i:9.4f}")
                legendticks.append(llabel)

            zi.unshare_mask()
            zi[zi < minv] = minv
            zi[zi > maxv] = maxv

            # need to restore the mask:
            zi.mask = zimask

            # note use surf.min, not usesurf.min here ...
            notetxt = (
                "Note: map values are truncated from ["
                + str(surf.values.min())
                + ", "
                + str(surf.values.max())
                + "] "
                + "to interval ["
                + str(minvalue)
                + ", "
                + str(maxvalue)
                + "]"
            )

            self._fig.text(0.99, 0.02, notetxt, ha="right", va="center", fontsize=8)

        logger.debug("Legendticks: %s", legendticks)

        if minvalue is None:
            minvalue = usesurf.values.min()

        if maxvalue is None:
            maxvalue = usesurf.values.max()

        # this will override current instance colormap locally, and is
        # therefore reset afterwards
        keepcolor = self.colormap
        if colormap is not None:
            self.colormap = colormap

        levels = np.linspace(minvalue, maxvalue, self.contourlevels)
        logger.debug("Number of contour levels: %s", levels)

        plt.setp(self._ax.xaxis.get_majorticklabels(), rotation=xlabelrotation)

        logger.debug("Current colormap is %s, requested is %s", self.colormap, colormap)
        logger.debug("Current colormap name is %s", self.colormap.name)

        uselevels = levels if ma.std(zi) > 1e-07 else 1

        try:
            if logarithmic is False:
                locator = None
                ticks = legendticks
                im = self._ax.contourf(
                    xi, yi, zi, uselevels, locator=locator, cmap=self.colormap
                )

            else:
                logger.debug("use LogLocator")
                locator = ticker.LogLocator()
                ticks = None
                uselevels = None
                im = self._ax.contourf(xi, yi, zi, locator=locator, cmap=self.colormap)

            self._fig.colorbar(im, ticks=ticks)
        except ValueError as err:
            logger.warning("Could not make plot: %s", err)

        plt.gca().set_aspect("equal", adjustable="box")
        self.colormap = keepcolor

    def plot_faults(
        self,
        fpoly,
        idname="POLY_ID",
        color="k",
        edgecolor="k",
        alpha=0.7,
        linewidth=0.8,
    ):
        """Plot the faults.

        Args:
            fpoly (object): A XTGeo Polygons object
            idname (str): Name of column which has the faults ID
            color (c): Fill color model c according to Matplotlib_
            edgecolor (c): Edge color according to Matplotlib_
            alpha (float): Degree of opacity
            linewidth (float): Line width

        .. _Matplotlib: http://matplotlib.org/api/colors_api.html
        """
        aff = fpoly.get_dataframe(copy=False).groupby(idname)

        for name, _group in aff:
            # make a dataframe sorted on faults (groupname)
            myfault = aff.get_group(name)

            # make a list [(X,Y) ...];
            af = list(zip(myfault["X_UTME"].values, myfault["Y_UTMN"].values))

            px = mplp.Polygon(af, alpha=alpha, color=color, ec=edgecolor, lw=linewidth)

            if px.get_closed():
                self._ax.add_artist(px)
            else:
                IOError(f"A polygon is not closed: {px}")

    def plot_polygons(self, fpoly, idname="POLY_ID", color="k", linewidth=0.8):
        """Plot a polygons instance.

        Args:
            fpoly (object): A XTGeo Polygons object
            idname (str): Name of column which has the faults ID
            color (c): Line color model c according to Matplotlib_
            linewidth (float): Line width

        .. _Matplotlib: http://matplotlib.org/api/colors_api.html
        """
        aff = fpoly.get_dataframe(copy=False).groupby(idname)

        for _name, group in aff:
            # make a dataframe sorted on groupname
            pname = fpoly.name

            xarr = group[fpoly.xname].values
            yarr = group[fpoly.yname].values

            self._ax.plot(xarr, yarr, label=pname, lw=linewidth, color=color)
            self._ax.legend()

    def plot_points(self, points):
        """Plot a points set on the map.

        This can be be useful e.g. for plotting the underlying point set
        that makes a gridded map.

        Args:
            points (Points): A XTGeo Points object X Y VALUE

        """
        # This function is "in prep"

        dataframe = points.get_dataframe(copy=False)

        self._ax.scatter(
            dataframe["X_UTME"].values, dataframe["Y_UTMN"].values, marker="x"
        )

    def plot_wells(self, wells):
        """Plot wells on the map.

        Args:
            wells (Wells): A XTGeo Wells object (contains a number of Well
                instances).

        """
        for well in wells.wells:
            dataframe = well.get_dataframe(copy=False)

            xval = dataframe["X_UTME"].values
            yval = dataframe["Y_UTMN"].values
            self._ax.plot(xval, yval)
            self._ax.annotate(well.name, xy=(xval[-1], yval[-1]))



"""Module for testing stand-alone scripts and/or entrypoints functions."""

import pytest
import xtgeo

from xtgeoviz import xsectplot
from xtgeoviz.frontends._xsectplotting_config import config_defaults, data_merge
from xtgeoviz.frontends.xsectplotting import _Xsections, _XsectSettings

# testdata (xtgeo-testdata), relative to testdir
WELLSET1 = "wells/drogon/1"
SURFACESET1 = "surfaces/drogon/1"
OUTLINE = "polygons/reek/1/closedpoly1.pol"
CUBE = "cubes/drogon/ampl_local_a4.segy"


def test_xsectplotting_config_defaults():
    """Test _xsectplottings_config() config_defaults method."""

    yamlresult = config_defaults()
    assert isinstance(yamlresult, str)

    dictresult = config_defaults(as_yaml=False)
    assert isinstance(dictresult, dict)
    assert dictresult["input"]["wells"]["zonelog"] == "ZONELOG"


def test_xsectplotting_config_dict_merge_ok():
    """Test _xsectplottings_config() data_merge method."""

    userinput = {
        "input": {
            "wells": {"zonelog": "MYZONELOG"},
        }
    }

    dictresult = config_defaults(as_yaml=False)
    update = data_merge(dictresult, userinput)
    assert update["input"]["wells"]["zonelog"] != "ZONELOG"
    assert update["input"]["wells"]["zonelog"] == "MYZONELOG"
    assert update["plotsettings"]["surfaces"]["contacts"]["colors"] == "rainbow"


def test_xsectplotting_config_dict_merge_shall_fail():
    """Test _xsectplottings_config() data_merge method shall fail due to invalid."""

    userinput = {
        "input": {
            "invalid_word": {"zonelog": "MYZONELOG"},
        }
    }

    dictresult = config_defaults(as_yaml=False)
    with pytest.raises(ValueError, match="Input contains invalid fields: invalid_word"):
        _ = data_merge(dictresult, userinput)


def test_xsects_settings_class():
    """Test the internal XsectsSettings dataclass class."""

    xsec = _XsectSettings()

    assert xsec.design_dpi == 100
    assert xsec.wells_zonelog_zoneshift == 0


def test_xsects_input_load_wells(testdir):
    """Test loading wells for xsect."""
    inputs = {
        "wells": {
            "folder": str(testdir / WELLSET1),
            "wildcard": "55*",
            "zonelog": "Zone",
        }
    }

    xapp = _Xsections(inputdata=inputs)
    xapp.load_wells()
    wellnames = [well.name for well in xapp.wells["wlist"]]
    for wname in ["55_33-1", "55_33-2", "55_33-3", "55_33-A-1", "55_33-A-2"]:
        assert wname in wellnames


def test_xsects_input_load_surfaces(testdir):
    """Test loading wells for xsect."""
    inputs = {
        "surfaces": {
            "primary": {
                "folder": str(testdir / SURFACESET1),
                "wildcard": "*",
                "names": ["TopVolantis", "TopTherys", "TopVolon", "BaseVolantis"],
            },
        },
    }

    xapp = _Xsections(inputdata=inputs)
    xapp.load_surfaces()
    surfnames = [surf.name for surf in xapp.surfaces["primary"]]
    for sname in ["TopTherys", "BaseVolantis"]:
        assert sname in surfnames


def test_xsectplot_function(testdir, tmp_path):
    """Make plots using python input."""

    inputs = {
        "wells": {
            "folder": str(testdir / WELLSET1),
            "wildcard": "55_33-A-4*",
            "zonelog": "Zone",
        },
        "surfaces": {
            "primary": {
                "folder": str(testdir / SURFACESET1),
                "wildcard": "*",
                "names": ["TopVolantis", "TopTherys", "TopVolon", "BaseVolantis"],
            },
        },
        "outline": str(testdir / OUTLINE),
    }
    psettings = {
        "design": {
            "zrange": [1550, 1750],
        },
    }

    outputs = {
        "plotfolder": str(tmp_path),
        "format": "png",
    }

    # call the top plotter routine
    xsectplot(inputdata=inputs, plotsettings=psettings, output=outputs)
    myplot = tmp_path / "55_33-A-4.png"
    assert myplot.is_file()


def test_xsectplot_function_input_objectlists(testdir, tmp_path):
    """Make plots using python input, where wells and surfaces are preloaded."""

    wells = ["55_33-A-1.rmswell", "55_33-A-4.rmswell"]
    surfaces = [
        "01_topvolantis.gri",
        "02_toptherys.gri",
        "03_topvolon.gri",
        "04_basevolantis.gri",
    ]

    wlist = []
    for well in wells:
        wlist.append(xtgeo.well_from_file(testdir / WELLSET1 / well))

    slist = []
    for surf in surfaces:
        slist.append(xtgeo.surface_from_file(testdir / SURFACESET1 / surf))

    inputs = {
        "wells": {
            "objects": wlist,
            "zonelog": "Zone",
        },
        "surfaces": {
            "primary": {
                "objects": slist,
                "names": ["TopVolantis", "TopTherys", "TopVolon", "BaseVolantis"],
            },
        },
    }
    psettings = {
        "design": {
            "zrange": [1550, 1750],
        },
        "wells": {
            "zonelog": {
                "zoneshift": 0,
            }
        },
    }

    outputs = {
        "plotfolder": str(tmp_path),
        "format": "png",
    }

    # call the top plotter routine
    xsectplot(inputdata=inputs, plotsettings=psettings, output=outputs)
    myplot = tmp_path / "55_33-A-4.png"
    assert myplot.is_file()


def test_xsectplot_function_include_cube(testdir, tmp_path):
    """Make plots using python input, where wells and surfaces are preloaded."""

    wells = ["55_33-A-4.rmswell"]
    surfaces = [
        "01_topvolantis.gri",
        "02_toptherys.gri",
        "03_topvolon.gri",
        "04_basevolantis.gri",
    ]

    wlist = []
    for well in wells:
        wlist.append(xtgeo.well_from_file(testdir / WELLSET1 / well))

    slist = []
    for surf in surfaces:
        slist.append(xtgeo.surface_from_file(testdir / SURFACESET1 / surf))

    cube = xtgeo.cube_from_file(testdir / CUBE)

    inputs = {
        "wells": {
            "objects": wlist,
            "zonelog": "Zone",
        },
        "surfaces": {
            "primary": {
                "objects": slist,
                "names": ["TopVolantis", "TopTherys", "TopVolon", "BaseVolantis"],
            },
        },
        "cube": cube,
    }
    psettings = {
        "design": {
            "zrange": {
                "default": [1000, 2000],
                "55_33.*": [1450, 1750],  # use a python regex for multiple wells
            },
        },
        "wells": {
            "zonelog": {
                "zoneshift": 0,
            }
        },
        "surfaces": {
            "primary": {
                "fill": False,
            },
        },
        "cube": {
            "range": [-0.33, 0.33],
            "sampling": "trilinear",
        },
    }

    outputs = {
        "plotfolder": str(tmp_path),
        "format": "png",
    }

    # call the top plotter routine
    xsectplot(inputdata=inputs, plotsettings=psettings, output=outputs)
    myplot = tmp_path / "55_33-A-4.png"
    assert myplot.is_file()



import pathlib

import xtgeoviz.plot._colortables as ct

TPATH = pathlib.Path("../xtgeo-testdata")


def test_readfromfile():
    """Read color table from RMS file."""
    cfile = TPATH / "etc/colortables/colfacies.txt"
    ctable = ct.colorsfromfile(cfile)
    assert ctable[5] == (0.49019608, 0.38431373, 0.05882353)


def test_xtgeo_colors():
    """Read the XTGeo color table."""
    ctable = ct.xtgeocolors()
    assert ctable[5] == (0.000, 1.000, 1.000)



"""Test the baseplot module and class."""

import matplotlib
import pytest
from matplotlib.testing.decorators import image_comparison

from xtgeoviz.plot.baseplot import BasePlot


@pytest.mark.skip(reason="Graphical compare testing immature")
@image_comparison(baseline_images=["canvas"], remove_text=False, extensions=["png"])
def test_canvas():
    """Test the canvas function."""

    myplot = BasePlot()
    myplot.canvas(title="Hello canvas", subtitle="Hello Subtitle", infotext="Some info")


def test_colormap():
    """Test colormap functions."""

    myplot = BasePlot()

    cmap = myplot.colormap
    colors = myplot.get_colormap_as_table()

    assert colors[1] == pytest.approx((0.26851, 0.009605, 0.335427, 1.0))
    assert cmap.name == "viridis"
    assert isinstance(cmap, matplotlib.colors.LinearSegmentedColormap)

    xtgeomap = myplot.define_any_colormap("xtgeo")
    myplot.colormap = xtgeomap
    assert isinstance(xtgeomap, matplotlib.colors.LinearSegmentedColormap)
    assert myplot.get_colormap_as_table()[1] == pytest.approx((0.0, 0.0, 0.0, 1))
    assert myplot.get_colormap_as_table()[8] == pytest.approx((0.8, 0.196, 0.6, 1.0))
    assert xtgeomap.name == "xtgeo"



"""Various helper functions for xsectplotting."""

from __future__ import annotations

import logging
import pathlib

import xtgeo

logger = logging.getLogger(__name__)


def load_wells(self):
    """Load wells from files, return as XTGeo well objects."""
    wlist = []

    if self.config["input"]["wells"]["objects"]:
        wlist = self.config["input"]["wells"]["objects"]

    else:
        wellfolder = pathlib.Path(self.config["input"]["wells"]["folder"])
        wcard = self.config["input"]["wells"]["wildcard"]

        for well in sorted(wellfolder.glob(wcard)):
            logger.info("Read well file: %s", well)
            wobj = xtgeo.well_from_file(
                well,
                zonelogname=self.config["input"]["wells"]["zonelog"],
                strict=False,
            )
            wlist.append(wobj)

        if not wlist:
            raise SystemExit("Cannot plot, no wells as input")

    self.wells["wlist"] = wlist

    self.wells["zonelog"] = self.config["input"]["wells"]["zonelog"]

    self.wells["facieslog"] = None
    if self.config["input"]["wells"]["facieslog"]:
        self.wells["facieslog"] = self.config["input"]["wells"]["facieslog"]

    self.wells["perflog"] = None
    if self.config["input"]["wells"]["perflog"]:
        self.wells["perflog"] = self.config["input"]["wells"]["perflog"]


def _load_surfaces_generic(self, case="primary"):
    """Load surfaces from files or grid, return as XTGeo RegularSurface objects.
    Note that the return is a dictionary with several possible surface
    sets: 'primary', 'secondary', 'tertiary', 'contacts'
    ::
    sdict:
       primary:
         [surfaceinstance1, surfaceinstance2, ...]
       contacts:
         [surfacieinstancex, surfacieinstancey, ...]
    Note the surface.name attribute is used to store the display surface name.
    """
    config = self.config
    sdict = {}
    names = config["input"]["surfaces"][case]["names"]

    if config["input"]["surfaces"][case]["objects"]:
        sdict = config["input"]["surfaces"][case]["objects"]

    else:
        surffolder = None
        wcard = None
        if "folder" in config["input"]["surfaces"][case]:
            actualfolder = config["input"]["surfaces"][case]["folder"]
            if actualfolder:
                surffolder = pathlib.Path(actualfolder)
        if "wildcard" in config["input"]["surfaces"][case]:
            wcard = config["input"]["surfaces"][case]["wildcard"]
            if not wcard:
                wcard = ""

        sobjects = []
        if surffolder and wcard:
            for inum, surf in enumerate(sorted(surffolder.glob(wcard))):
                logger.info("Read surface file: %s", surf)
                sobjects.append(xtgeo.surface_from_file(surf))

        sdict = sobjects

    for inum, srf in enumerate(sdict):
        if not isinstance(srf, xtgeo.RegularSurface):
            raise RuntimeError("Surface object are not XTGeo objects")
        if names:
            srf.name = names[inum]
        else:
            srf.name = "Surface_" + str(inum)

    self.surfaces[case] = sdict


def load_surfaces(self):
    for casename in ["primary", "secondary", "contacts"]:
        _load_surfaces_generic(self, case=casename)


def load_outline(self):
    """Load field outline from file, return as XTGeo Polygons objects."""

    outline = self.config["input"]["outline"]

    if isinstance(outline, str):
        logger.info("Read outline: %s", outline)
        self.outline = xtgeo.polygons_from_file(outline)
    elif isinstance(outline, xtgeo.Polygons):
        self.outline = outline


def load_cube(self):
    """Apply a current cube or load a cube, both as XTGeo Cube"""

    cube = self.config["input"]["cube"]

    if cube and isinstance(cube, str):
        logger.info("Reading cube: %s", cube)
        self.cube = xtgeo.cube_from_file(cube)
        logger.info("Reading cube done")
    elif cube and isinstance(cube, xtgeo.Cube):
        self.cube = cube
        logger.info("Apply an existing Cube instance: %s", type(cube))
    else:
        self.cube = None


def load_grid(self):
    """Load a grid with property"""

    gfile = self.config["input"]["grid"]["geometry"]
    pfile = self.config["input"]["grid"]["property"]["file"]
    pname = self.config["input"]["grid"]["property"]["name"]
    pdate = self.config["input"]["grid"]["property"]["date"]

    read_grid = True
    if gfile == self.config["input"]["surfaces"]["fromgrid"]:
        # skip reading grid once more
        read_grid = False

    if gfile and pfile:
        if read_grid:
            logger.info("Reading grid geometry: %s", gfile)
            self.grid = xtgeo.grid_from_file(gfile)
            logger.info("Reading grid geometry done")
        logger.info("Reading grid property: %s", pfile)
        self.gridproperty = xtgeo.gridproperty_from_file(pfile, name=pname, date=pdate)
        logger.info("Reading grid property done")



"""Module for 3D Grid slice plots, using matplotlib."""

from __future__ import annotations

import logging

import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon

from .baseplot import BasePlot

logger = logging.getLogger(__name__)


class Grid3DSlice(BasePlot):
    """Class for plotting a row, a column, or a layer, using matplotlib."""

    def __init__(self):
        """Construct an instance for a Grid3DSlice object."""
        super().__init__()

        self._wells = None
        self._surface = None
        self._tight = False

        self._wfence = None
        self._legendtitle = "Map"

        self._colormap = "rainbow"
        self._linecolor = "black"
        self._clist = None
        self._prop = None
        self._grid = None
        self._geomlist = None
        self._index = 1
        self._actnum = None
        self._window = None
        self._active = True
        self._minvalue = None
        self._maxvalue = None

    def plot_gridslice(
        self,
        grid,
        prop=None,
        mode="layer",
        minvalue=None,
        maxvalue=None,
        colormap=None,
        linecolor="black",
        index=1,
        window=None,
        activeonly=True,
    ):
        """Plot a row slice, column slice or layer slice of a grid.

        Args:
            grid (Grid): The XTGeo grid object
            prop (GridProperty, optional): The XTGeo grid property object
            mode (str): Choose between 'column', 'row', 'layer' (default)
            minvalue (float): Minimum level color scale (default: from data)
            maxvalue (float): Maximum level color scale (default: from data)
            index (int): Index to plot e.g layer number if layer slice (first=1)
            colormap: Color map to use for cells, e.g. 'rainbow' or an rmscol file
            linecolor (str or tuple): Color of grid lines (black/white/grey
                or a tuple with 4 numbers on valid matplotlib format)
            activeonly (bool): If only use active cells
            window (str): Some window

        """
        self._index = index
        if colormap is not None:
            self._colormap = colormap

        self._linecolor = linecolor
        if not isinstance(linecolor, tuple) and linecolor not in (
            "black",
            "grey",
            "white",
        ):
            raise ValueError("Value of linecolor is invalid")

        self._clist = grid.get_xyz_corners()  # get XYZ for each corner, 24 arrays
        self._grid = grid
        self._prop = prop
        self._window = window

        if self._geomlist is None:
            # returns (xori, yori, zori, xmin, xmax, ymin, ymax, zmin, zmax,...)
            self._geomlist = grid.get_geometrics(allcells=True, cellcenter=False)

        self._active = activeonly

        self._minvalue = minvalue
        self._maxvalue = maxvalue

        if mode in ("column", "row"):
            pass  # self._plot_row()
        else:
            self._plot_layer()

    def _plot_layer(self):
        xyc, ibn = self._grid.get_layer_slice(self._index, activeonly=self._active)

        xval = xyc[:, :, 0]
        yval = xyc[:, :, 1]

        xvmin = xval.min()
        xvmax = xval.max()
        yvmin = yval.min()
        yvmax = yval.max()

        if self._window is None:
            xmin = xvmin - 0.05 * (abs(xvmax - xvmin))
            xmax = xvmax + 0.05 * (abs(xvmax - xvmin))
            ymin = yvmin - 0.05 * (abs(yvmax - yvmin))
            ymax = yvmax + 0.05 * (abs(yvmax - yvmin))
        else:
            xmin, xmax, ymin, ymax = self._window

        patches = []

        for pos in range(len(ibn)):
            nppol = xyc[pos, :, :]
            if nppol.mean() > 0.0:
                polygon = Polygon(nppol)
                patches.append(polygon)

        patchcoll = PatchCollection(
            patches, edgecolors=(self._linecolor,), cmap=self.colormap
        )

        if self._prop:
            pvalues = self._prop.values
            pvalues = pvalues[:, :, self._index - 1]
            pvalues = pvalues[~pvalues.mask]

            patchcoll.set_array(pvalues)

            pmin = self._minvalue
            if self._minvalue is None:
                pmin = pvalues.min()

            pmax = self._maxvalue
            if self._maxvalue is None:
                pmax = pvalues.max()

            patchcoll.set_clim([pmin, pmax])

        im = self._ax.add_collection(patchcoll)
        self._ax.set_xlim((xmin, xmax))
        self._ax.set_ylim((ymin, ymax))
        self._fig.colorbar(im)

        plt.gca().set_aspect("equal", adjustable="box")



"""Module for quickplot of different xtgeo data types.

Typically::

    import xtgeo
    from xtgeoviz import quickplot

    surf = xtgeo.surface_from_file("some.gri")

    quickplot(surf)

"""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Optional, Union

import xtgeo

from .xtmap import Map

if TYPE_CHECKING:
    import pathlib


def quickplot(xobj, **kwargs):
    """A common entry point for quickplot."""
    if isinstance(xobj, xtgeo.RegularSurface):
        quickplot_regularsurface(xobj, **kwargs)


def quickplot_regularsurface(
    regsurf: xtgeo.RegularSurface,
    filename: Optional[Union[str, pathlib.Path]] = None,
    title: Optional[str] = "QuickPlot for Surface",
    subtitle: Optional[str] = None,
    infotext: Optional[str] = None,
    minmax: tuple[int | float | None, int | float | None] = (None, None),
    xlabelrotation: Optional[float] = None,
    colormap: Optional[str] = "rainbow",
    faults: Optional[xtgeo.Polygons] = None,
    logarithmic: Optional[bool] = False,
):
    """Quickplot regularsurface."""

    ncount = regsurf.values.count()
    if ncount < 5:
        warnings.warn(
            f"None or too few map nodes for plotting. Skip output {filename}!",
            UserWarning,
        )
        return

    mymap = Map()
    mymap.canvas(title=title, subtitle=subtitle, infotext=infotext)

    minvalue, maxvalue = minmax
    mymap.colormap = colormap

    mymap.plot_surface(
        regsurf,
        minvalue=minvalue,
        maxvalue=maxvalue,
        xlabelrotation=xlabelrotation,
        logarithmic=logarithmic,
    )
    if faults:
        mymap.plot_faults(faults["faults"])

    if filename is None:
        mymap.show()
    else:
        mymap.savefig(filename)



"""Making cross sections (plots), primarely wells."""

from __future__ import annotations

import argparse
import json
import logging
import os.path
import sys
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Union

import yaml

from . import (
    _xsectplotting_config as _cfg,
    _xsectplotting_load as _load,
    _xsectplotting_plotting as _plt,
)

APPNAME = "xtgeoviz.xsectplot"

logger = logging.getLogger(__name__)


def _prettyprint(anydict):
    try:
        print(json.dumps(anydict, indent=4, ensure_ascii=False))
    except TypeError:
        print("Cannot serialize to JSON...")


@dataclass
class _XsectSettings:
    """Class for storing plot settings for the plots"""

    title: str = "Generic Title"

    general_engine: str = "matplotlib"

    design_style: int = 1
    design_zrange: Union[tuple, list, dict] = (1000, 2000)
    design_gridlines: str = "both"
    design_subtitle: Optional[str] = None
    design_legendsize: int = 6
    design_legends: bool = True
    design_dpi: int = 100

    wellmap_otherwells: bool = False
    wellmap_expand: int = 2

    surf_primary_colors: str = "random40"
    surf_primary_fill: bool = True

    surf_contacts_colors: str = "rainbow"

    # surfaces secondary:
    surf_secondary_colors: str = "xtgeo"
    surf_secondary_legend: str = "?"

    cube_colors: str = "seismic"
    cube_range: Any = None
    cube_interpolation: str = "gaussian"
    cube_alpha: float = 0.8
    cube_sampling: str = "nearest"

    wells_zonelog_colors: str = "random40"
    wells_zonelog_zoneshift: int = 0
    wells_zonelog_colordict: dict = field(default_factory=dict, init=False)
    wells_facieslog_colors: str = "xtgeo"
    wells_facieslog_colordict: dict = field(default_factory=dict, init=False)
    wells_perflog_colors: str = "xtgeo"
    wells_perflog_colordict: dict = field(default_factory=dict, init=False)
    wells_wellcrossings_show: bool = False
    wells_wellcrossings_sampling: int = 20
    wells_wellcrossings_wfilter: int = 5

    output_plotfolder: str = "/tmp"
    output_format: str = "svg"
    output_prefix: str = ""
    output_pdfjoin: bool = False
    output_cleanup: bool = False

    def __post_init__(self):
        logger.info("Initilize class: %s", __class__.__name__)

    def update_plotsettings(self, config):
        logger.info("Update plotsettings")
        _cfg.update_plotsettings(self, config)


@dataclass
class _Xsections:
    """Private class for XSections plotting"""

    args: Optional[list] = None
    inputdata: dict = field(default_factory=dict)
    psettings: dict = field(default_factory=dict)
    output: dict = field(default_factory=dict)
    verbosity: str | None = None

    config: dict = field(default_factory=dict, init=False)  # resulting YAML config
    polylines: list = field(default_factory=list, init=False)
    multiwells: list = field(default_factory=list, init=False)  # list of multiwells
    wells: dict = field(default_factory=dict, init=False)
    cube: Any = field(default=None, init=False)
    grid: Any = field(default=None, init=False)
    gridproperty: Any = field(default=None, init=False)
    outline: Any = field(default=None, init=False)
    contacts: Any = field(default=None, init=False)
    surfaces: dict = field(default_factory=dict, init=False)

    def __post_init__(self):
        self.args = None  # command line args
        self.plotsettings = _XsectSettings()  # plotsettings

        if not self.inputdata:
            self.parse_args()
        else:
            self.parse_dicts()

    def parse_args(self):
        """Parse the command line arguments."""
        parser = argparse.ArgumentParser(
            description="Make beautiful cross sections in seconds",
            usage="xsections --config some.yaml ... ",
        )

        parser.add_argument(
            "-c", "--config", dest="config", type=str, help="Config file on YAML format"
        )

        if self.args is None:
            self.args = sys.argv[1:]

        if len(self.args) < 2:
            parser.print_help()
            print("QUIT")
            sys.exit(0)

        self.args = parser.parse_args(self.args)

    def set_logger_verbosity(self):
        """Use config setting to set eventual logging output."""

        actual_verbosity = "normal"
        if self.verbosity:
            actual_verbosity = self.verbosity
        else:
            actual_verbosity = self.config["verbosity"]

        if actual_verbosity == "silent":
            logging.basicConfig(
                level=logging.WARNING,
                stream=sys.stdout,
                format="%(asctime)s %(levelname)s:%(message)s",
            )
            for handler in logging.root.handlers:
                handler.addFilter(logging.Filter("xtgeoviz"))

        elif actual_verbosity == "info":
            logging.basicConfig(
                level=logging.INFO,
                stream=sys.stdout,
                format="%(asctime)s %(levelname)s:%(message)s",
            )
            for handler in logging.root.handlers:
                handler.removeFilter(logging.Filter("xtgeoviz"))

        else:  # normal
            logging.basicConfig(
                level=logging.INFO,
                stream=sys.stdout,
                format=">> %(message)s",
            )
            for handler in logging.root.handlers:
                handler.addFilter(logging.Filter("xtgeoviz"))

    def parse_config(self):
        """Read config from YAML and oveeride with defaults."""

        args = self.args
        inputfile = args.config

        ydefaults = _cfg.config_defaults()

        dconfig = yaml.safe_load(ydefaults)

        if not os.path.isfile(inputfile):
            raise OSError(f"No such config file exists: {inputfile}")

        with open(inputfile, encoding="utf-8") as stream:
            config = yaml.safe_load(stream)

        # merge config with defaults:
        merged = _cfg.data_merge(dconfig, config)

        self.config = merged
        self.set_logger_verbosity()

        logger.info("Input config YAML file <%s> is read...", inputfile)

    def parse_dicts(self):
        """Parse settings from dicts given as arguments, and merge with defaults."""
        default = _cfg.config_defaults(as_yaml=False)
        # _prettyprint(default)
        update = {
            "input": self.inputdata,
            "plotsettings": self.psettings,
            "output": self.output,
        }
        self.config = _cfg.data_merge(default, update)
        self.set_logger_verbosity()
        # _prettyprint(self.config)

    def load_wells(self):
        """Load wells as a list of XTGeo Well() instances"""

        _load.load_wells(self)

    def load_surfaces(self):
        """Load surfaces as a dict of lists of XTGeo Surface() instances"""

        _load.load_surfaces(self)

    def load_cube(self):
        """Load cube to plot as backdrop, a XTGeo Cube() instances"""

        _load.load_cube(self)

    def load_grid(self):
        """Load grid with property to plot as backdrop, XTGeo Grid() + GridProperty
        instances.
        """

        _load.load_grid(self)

    def load_outline(self):
        """Load map outline to plot as info"""

        _load.load_outline(self)

    def config_complete(self):
        """improve config, better defaults, and some smart settings based on input"""

        _cfg.config_complete(self)  # config, wells, surfs["primary"])

    def plot(self):
        """Do the actual plotting"""

        _plt.plotting(self)


# ======================================================================================
# MAIN
# ======================================================================================


def xsectplot(
    args: Optional[Dict] = None,
    inputdata: Optional[Dict] = None,
    plotsettings: Optional[Dict] = None,
    output: Optional[Dict] = None,
    verbosity: Optional[str] = None,
):
    """Frontend function for plotting xsections.

    The input can either be in form of command line args parsing a YAML file with
    additional command line args that overides, or it can be dicts in form of inputdata,
    plotsettings and output that will be combined with defaults.

    If the input key is not None, then any args will be ignored.

    Args:
        args: Command line args
        input: A dictionary spesifying minimum required input
        plotsettings: A dictionary spesifying minimum required plotsettings
        output: A dictionary spesyfying minimum required output settings
        verbosity: Default None means that it is set from the config; otherwise a
            string here will override: "normal" for normal output, "info" for using
            logging for all modules, and similarly "debug"
    """

    app = _Xsections(args, inputdata, plotsettings, output, verbosity)

    # load what to xsect (and show):
    app.load_wells()
    app.load_surfaces()
    app.load_cube()
    app.load_outline()

    app.config_complete()

    app.plotsettings.update_plotsettings(app.config)

    app.plot()


if __name__ == "__main__":
    xsectplot()
